#' @description CNVmap detects signatures of copy-number variants (CNVs) in genetic
#'    segregation data produced for instance for linkage mapping. It uses the information
#'    of apparent heterozygotes arising when a given genetic marker targets two
#'    divergent copies of a duplicated region. Based on these types of situations, CNVmap
#'    extracts different subsamples of individuals from the mapping population and then
#'    produces the allele frequency patterns along the
#'    genome from those subsamples. It then automatically analyses these patterns
#'    and infers the type of CNV involved, as well as
#'    the genetic positions of the associated copies.
"_PACKAGE"





#' @title              CNVdetect
#' 
#' @description Analyzes a genetic segregation data set to detect signatures
#' of copy-number variants (CNVs). Characterizes the type of event and maps the new
#' loci discovered.\cr
#' This function requires the following data files in the working directory:\cr
#' (1) File <popName>.map tab-separated text file without headers, with\cr
#'  - First column character string: marker names. \cr
#'  - Second column character string: chromosome name (numeric or alphanumeric).\cr
#'  - Third column numeric: genetic map position in centiMorgans.\cr
#' (2) File <popName>.seg tab-separated text file without headers containing:\cr
#'  - A first line with all names of individuals in the population,\cr
#'  - Then one line per marker with marker name as first column and genotype 
#'    calls ("A", "B", "H", or "-") of each individual of the population in 
#'    subsequent columns, with the order of individuals corresponding to that
#'    of the first line\cr
#' Alternatively, segregation data may be provided as MapMaker files (.raw file)
#' or as SNP genotype files (.snp file). See details of these formats
#' in function load_data()
#'
#' @param popName character string: name of the data set to be analyzed. This 
#' name must be used as prefix for the data files 
#' @param mappingPanelType character string: may be one of the following: "DH" for
#' doubled-haploids, "RIL" for recombinant inbred lines, or "IRIL" for intermated
#' recombinant inbred lines having undergone a number of random mating generations 
#' before selfing or sib-mating
#' @param parents vector of 2 character strings: vector of the names of the two
#' parents, as they will appear in the outputs
#' @param nbSelfingGenerationsRils numeric: number of generations of selfing
#' or sib-mating performed for recombinant inbred lines
#' @param nbIntermatingGenerations numeric: number of panmictic generations 
#' before selfing in the case of IRILs
#' @param nuGamma numeric: strength of crossover interference, expressed
#' as the parameter 'nu' of the Gamma model. Used to incorporate interference
#' in simulations of segregation data. nu=1 means no interference, nu>1 means
#' positive crossover interference
#' @param maxPercentGenot numeric: maximum frequency (in percentage) of any
#' genotype (A, B, or H) allowed for an individual to be kept for the analyses
#' @param quantThresHetMend numeric: threshold for the maximum quantile of 
#' heterozygous calls for a marker to be used in allele frequency analyses
#' @param quantThresMissMend numeric: threshold for the maximum quantile of 
#' missing data calls for a marker to be used in allele frequency analyses
#' @param minMafMend numeric: minimum value of Minor Allele Frequency for
#' a marker to be used in allele frequency analyses
#' @param analyzeMarkersWithoutHetero logical: if TRUE, markers with anomalously 
#' high rates of missing data will also be considered as candidates
#' for being associated to CNV events (otherwise only those with anomalously 
#' high rates of heterozygote calls will be candidates)
#' @param quantThresHetCand numeric: threshold for the maximum quantile of 
#' heterozygous calls for a marker to be considered as candidates
#' for being associated to CNV events
#' @param quantThresMissCand numeric: threshold for the maximum quantile of 
#' missing data for a marker to be considered as candidates for being
#' associated to CNV events
#' @param minNbMrkPerTarget numeric: minimum number of markers required
#' within a peak for the peak to be considered valid
#' @param maxGenDistToFusePeaks numeric: maximum genetic distance
#' in centiMorgans below which two detected peaks will be fused
#' @param winSizeToFitPeakShape numeric: genetic distance in
#' centiMorgans on each side of a peak, where linear regression
#' analysis will be performed to check the shape of the peak
#' @param minSlopePeak numeric: minimum slope of the regression line
#' fitted on either side of a putative peak
#' @param toleranceFreq numeric: maximum allele frequency considered
#' as zero for peak detection (or symmetrically close to one)
#' @param minR2peak numeric: minimum value of the regression R-square
#' on either side of a peak for valid detection
#' @param simulateProfiles logical: if TRUE, for each candidate marker 
#' which leads to the identification of a CNV event, segregation data
#' will be simulated under the hypothesis of the identified type of
#' CNV, and used to plot allele frequency profiles which can be compared
#' to the ones obtained with the experimental data. Such simulation
#' takes some additional time, but provides an important means to assess
#' a posteriori the quality of the automatic identification of the CNV
#' type performed by the software
#' @param plotCNVstrong logical: if TRUE, generates plots for marker candidates
#' which lead to the identification of a CNV with strong signature (that is
#' based on at least one peak of non-missing data per locus). Useful 
#' to visually assess the validity of the automatic identification 
#' of CNVs, and check for possible false positives
#' @param plotCNVweak logical: if TRUE, generates plots for marker candidates
#' which lead to the identification of a CNV with weak signature (that is 
#' based on only one peak of missing data). This may increase the size of
#' output files without producing new true positives
#' @param plotUnclassified logical: if TRUE, generates plots for events
#' in which peaks have been detected, but did not match any of the
#' CNV signatures considered in the software. Such plots may be used 
#' to estimate by eye the level of false negatives, or to discover
#' new types of events which were not considered in the software but
#' may be of interest
#' @param plotSingleLocus logical: if TRUE, plots allele frequency profiles
#' for candidate markers for which only the peaks corresponding to the
#' expected signature of a single locus were detected. May be useful to 
#' check for possible false negatives but may produce huge files
#' @param customCandidateList vector of character strings. List of names
#' of candidate markers to be analyzed. If NULL, then all candidates will 
#' be kept. If the list contains markers that are not candidates, they
#' will not be analyzed.
#' @param numberShufflesForPvalue numeric: number of profiles simulated under
#' the hypothesis H0="single-locus" to calculate a p-value. Caution: may be long...
#' If equal to zero (default), then no P-value is calculated.
#' @param printDistribTypesShuffles logical: if TRUE outputs distribution
#' of CNV types from all shuffled data generated to calculate P-values
#' @param plotShuffles logical: if TRUE generates and saves profile plots 
#'for all shuffled data generated to calculate P-values. Caution: this will
#' generate one file for each candidate marker!
#' @param nbCpus numeric: number of CPUs used for parallelization of the
#' shuffles to compute the p-value for H0="single-locus". If nbCpu=1
#' then no parallelization, otherwise requires installing the package 
#' 'doParallel' which requires 'foreach' and 'iterators'
#' @examples CNVdetect(popName="GABI")
#' @export
#' 

CNVdetect = function(  popName ,
                       # PARAMETERS RELATED TO MAPPING POPULATION
                       mappingPanelType="DH",               # May only be "DH" or "RIL" or "IRIL"
                       parents = c("P1","P2"),
                       nbSelfingGenerationsRils = 6,        # number of selfing or sib-mating generations (only for RILs or IRILs)
                       nbIntermatingGenerations = 4,        # number of random intermating generations in (only for IRILs)
                       nuGamma = 3,
                       # PARAMETERS TO FILTER SUSPICIOUS INDIVIDUALS
                       maxPercentGenot = 90,                # Max percentage of any genotype (A, B, or H) to keep an individual
                       # PARAMETERS TO IDENTIFY NON-MENDELIAN MARKERS
                       quantThresHetMend = 70,              # Max quantile % threshold of Het freq for a marker to be Mendelian
                       quantThresMissMend = 70,  		    # Max quantile % threshold of Missing freq for a marker to be Mendelian
                       minMafMend = 0.15,					# Minimum of Minor Allele Frequency for a Mendelian marker
                       # PARAMETERS RELATED TO CNV DETECTION
                       analyzeMarkersWithoutHetero = TRUE,  # Considers also markers with no H calls but at least one missing call
                       quantThresHetCand = 93,				# Min quantile % threshold of Het for a marker to be Candidate
                       quantThresMissCand = 95,				# Min quantile % threshold of Missing for a marker to be Candidate
                       minNbMrkPerTarget = 2,				# Minimum number of markers within one target (freq=0 or 1)
                       maxGenDistToFusePeaks = 10,			# cM dist below which two detected peaks are fused
                       winSizeToFitPeakShape = 40,			# cM length on each side of peak, where shape will be fit to a line
                       minSlopePeak = -0.05,			    # Minimum slope of the regression line on either side of the peak
                       toleranceFreq = 0.05,				# maximum allele freq considered as zero (or symmetrically close to one)
                       minR2peak = 0.7,
                       simulateProfiles = TRUE,
                       plotCNVstrong = TRUE,                # if TRUE, plots profiles if a CNV event was identified
                       plotCNVweak = TRUE,		            # if TRUE, plots profiles if a CNV event was identified
                       plotUnclassified = TRUE,		   	    # if TRUE, plots profiles if there are peaks detected but no matching CNV hypothesis
                       plotSingleLocus = FALSE,		   	    # if TRUE, plots profiles if only a single locus was detected
                       customCandidateList = NULL,	        # if TRUE, plots profiles when only a single locus was identified
                       numberShufflesForPvalue = 100,       # if not zero, p-values will be calculated for H0="single locus"
                       printDistribTypesShuffles = TRUE,	# if TRUE outputs distribution of CNV types from all shuffles used for P-values
                       plotShuffles = FALSE,		        # if TRUE generates and saves profile plots for all shuffles used for P-values
                       nbCpus = 1 ) {						# if 1 does not use parallelization, otherwise requires installing 'doParallel'
            
	if (nbCpus > 1) {
		library(foreach)
		library(iterators)
		library(doParallel)
	}	
    singleLocusEvents <- c("A/B","A/B|Ah","A/B|Am","A/B|Bh","A/B|Bm")
    CNVweak <- c("A/B_A/-|HmHb", "A/B_-/A|HmHb", "A/B_B/-|HmHa", "A/B_-/B|HmHa")
    CNVstrong <- c("A/B_A/-", "A/B_A/-|Hm"
                 , "A/B_-/A", "A/B_-/A|Hm"
                 , "A/B_B/-", "A/B_B/-|Hm"
                 , "A/B_-/B", "A/B_-/B|Hm"
                 , "A/B_-/A_-/A", "A/B_-/A_A/-", "A/B_-/A_-/B", "A/B_-/A_B/-"
                 , "A/B_A/-_-/A", "A/B_A/-_A/-", "A/B_A/-_-/B", "A/B_A/-_B/-"
                 , "A/B_-/B_-/A", "A/B_-/B_A/-", "A/B_-/B_-/B", "A/B_-/B_B/-"
                 , "A/B_B/-_-/A", "A/B_B/-_A/-", "A/B_B/-_-/B", "A/B_B/-_B/-"  )
    
    datePrefix <- paste0(popName, "_", Sys.Date(),format(Sys.time(), "_%H-%M-%S"))
    logFileName <- paste0(datePrefix, "_", "CNVmap.log")
    segFileName <- paste0(popName, ".seg")
    snpFileName <- paste0(popName, ".snp")
    rawFileName <- paste0(popName, ".raw")
    mapFileName <- paste0(popName, ".map")
    listFileName <- paste0(datePrefix, "_CNV_List.txt")
    positionsFileName <- paste0(datePrefix, "_CNV_Positions.txt")
    statsFileName <- paste0(datePrefix, "_CNV_Stats.txt")
    profileSuffix <- "_Profiles.pdf"
    sink(logFileName, type="output", split=TRUE)
    
    cat ("
==================================================
==               CNVmap  v1.2                   ==
==================================================\n")
    cat(date()) ; cat("\n\n")
    cat(paste0("Data name: ", popName)) ; cat("\n\n")
    params <- list(
        popName = popName,
        mappingPanelType = mappingPanelType,
        parents = paste(parents, collapse = "-"),
        nbSelfingGenerationsRils = nbSelfingGenerationsRils,
        nbIntermatingGenerations = nbIntermatingGenerations,
        nuGamma = nuGamma,
        maxPercentGenot = maxPercentGenot,
        quantThresHetMend = quantThresHetMend,
        quantThresMissMend = quantThresMissMend,
        minMafMend = minMafMend,
        analyzeMarkersWithoutHetero = analyzeMarkersWithoutHetero,
        quantThresHetCand = quantThresHetCand,
        quantThresMissCand = quantThresMissCand,
        minNbMrkPerTarget = minNbMrkPerTarget,
        maxGenDistToFusePeaks = maxGenDistToFusePeaks,
        winSizeToFitPeakShape = winSizeToFitPeakShape,
        minSlopePeak = minSlopePeak,
        toleranceFreq = toleranceFreq,
        minR2peak = minR2peak,
        simulateProfiles = simulateProfiles,
        plotCNVstrong = plotCNVstrong,
        plotCNVweak = plotCNVweak,
        plotUnclassified = plotUnclassified,
        plotSingleLocus = plotSingleLocus,
        customCandidateList = customCandidateList,
        numberShufflesForPvalue = numberShufflesForPvalue,
        printDistribTypesShuffles = printDistribTypesShuffles,
        plotShuffles = plotShuffles,
        nbCpus = nbCpus
    )
    displayPars <- data.frame("."=as.matrix(params))
    cat("PARAMETER DETAILS:\n")
    print(displayPars)
    cat("\n")
    cat("List of Copy-Number Variation events to look for:\n")
    hypot <- get_hypotheses()
    hypotheses <- hypot$hyp
    err_funcs <- hypot$estimate_errors
    print(hypot$hypDisplay)
    cat("\n")
    firstStats <- data.frame(Item=names(params), Value=as.matrix(params))
    row.names(firstStats) <- c()
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="----------------------------------------", Value=0))
    
    ldRet <- load_data(mapFileName=mapFileName, segFileName=segFileName, rawFileName=rawFileName
                       , snpFileName=snpFileName, mappingPanelType=mappingPanelType)
    seg <- ldRet$seg  ;  map <- ldRet$map  ;  nbMrk1 <- dim(seg)[1]
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="Markers_Before_Filtering", Value=dim(seg)[1]))
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="Individuals_Before_Filtering", Value=dim(seg)[2]))
    
    res <- filter_markers_plants(map=map, seg=seg, maxPercentGenot=maxPercentGenot, minMAF=minMafMend)
    seg = res$seg  ;  map = res$map  ;  nbMrk2 <- dim(seg)[1]
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="Markers_Remaining_After_Filtering", Value=dim(seg)[1]))
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="Individuals_Remaining_After_Filtering", Value=dim(seg)[2]))
    
    chrs <- unique(map$chr)
    allChromLens <- rep(NA, length(chrs))
    for (iChr in 1:length(chrs)) allChromLens[iChr] <- max(map[map$chr==chrs[iChr],]$genPos)
    
    # Setting default thresholds for selecting candidate and Mendelian markers
    bonferroniCorrection <- sum(allChromLens)/100  # considers that markers are independent at more than 100 centiMorgan (value for DH)
    if (mappingPanelType == "RIL") bonferroniCorrection <- bonferroniCorrection * 2  #  number of effective meioses
    if (mappingPanelType == "IRIL") bonferroniCorrection <- bonferroniCorrection * (2 + nbIntermatingGenerations / 2)  #  number of effective meioses
    # bonferroniCorrection is the number of independent regions genome-wide for peak detection
    limitminNumIndivForCurve <- ceiling(log(4*bonferroniCorrection)/log(2)) 
    # For this number of H plants, the proba of discovering at least 1 spurious peak genome-wide is at most 0.5
    sortedNbHperMrk <- sort(apply(X=seg, MARGIN=1, FUN=function(x) sum(x=="H")))
    sortedNbMperMrk <- sort(apply(X=seg, MARGIN=1, FUN=function(x) sum(x=="-")))
    
    maxNumIndivHmend <- sortedNbHperMrk[floor(length(sortedNbHperMrk) * quantThresHetMend/100)]
    cat(paste0("Max number of H calls across indiv for a marker to be Mendelian set to ",maxNumIndivHmend)) ; cat("\n")
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="Max_Num_Indiv_H_for_Mend", Value=maxNumIndivHmend))
    maxNumIndivMmend <- sortedNbMperMrk[floor(length(sortedNbMperMrk) * quantThresMissMend/100)]
    cat(paste0("Max number of Missing calls across indiv for a marker to be Mendelian set to ",maxNumIndivMmend)) ; cat("\n")
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="Max_Num_Indiv_M_for_Mend", Value=maxNumIndivMmend))
    
    minNumIndivHcand <- sortedNbHperMrk[ceiling(length(sortedNbHperMrk) * quantThresHetCand/100)]
    minNumIndivHcand <- max(minNumIndivHcand, limitminNumIndivForCurve)
    cat(paste0("Min number of H calls across indiv for a marker to be Candidate set to ",minNumIndivHcand)) ; cat("\n")
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="Min_Num_Indiv_H_for_Cand", Value=minNumIndivHcand))
    minNumIndivMcand <- sortedNbMperMrk[ceiling(length(sortedNbMperMrk) * quantThresMissCand/100)]
    minNumIndivMcand <- max(minNumIndivMcand, limitminNumIndivForCurve)
    cat(paste0("Min number of Missing calls across indiv for a marker to be Candidate set to ",minNumIndivMcand)) ; cat("\n") ; cat("\n")
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="Min_Num_Indiv_M_for_Cand", Value=minNumIndivMcand))
    
    ret <- mrk_with_het_or_missing(seg=seg, map=map, minMAF=minMafMend, maxNumIndivHmend=maxNumIndivHmend, maxNumIndivMmend=maxNumIndivMmend
                                   , minNumIndivHcand=minNumIndivHcand, minNumIndivMcand=minNumIndivMcand 
                                   , selMrkWithHet=TRUE, selMrkWithMissing=analyzeMarkersWithoutHetero) 
    canMrkNames <- ret$canMrkNames  ;  seg <- ret$seg  ;  map <- ret$map  ;  nbMrk3 <- dim(seg)[1]
    nbMendelian <- dim(seg)[1] - length(canMrkNames)
    nbNonMendNonCan <- nbMrk2 - nbMrk3
    cat("\nCandidates:",length(canMrkNames), "    Mendelian Markers:", nbMendelian, "    Non-Mendelian non-candidates:", nbNonMendNonCan,"\n")
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="Candidate_Markers", Value=length(canMrkNames)))
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="Mendelian_Markers", Value=nbMendelian))
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="Non-Mendelian_Non-Candidate_Markers", Value=nbNonMendNonCan))
    firstStats <- rbind.data.frame(firstStats, data.frame(Item="----------------------------------------", Value=0))
    
    devList <- c(CNVstrong=NA, CNVweak=NA, Unclassified=NA, SingleLocus=NA) ; devNum <- 1
    if (plotCNVstrong)    { pdf(paste0(datePrefix, "_CNV_Strong", profileSuffix), 10, 1.5)  ; par(cex=0.4); devNum<-devNum+1; devList[1]<-devNum }
    if (plotCNVweak)      { pdf(paste0(datePrefix, "_CNV_Weak", profileSuffix), 10, 1.5)    ; par(cex=0.4); devNum<-devNum+1; devList[2]<-devNum }
    if (plotUnclassified) { pdf(paste0(datePrefix, "_Unclassified", profileSuffix), 10, 1.5); par(cex=0.4); devNum<-devNum+1; devList[3]<-devNum }
    if (plotSingleLocus)  { pdf(paste0(datePrefix, "_Single_Locus", profileSuffix), 10, 1.5); par(cex=0.4); devNum<-devNum+1; devList[4]<-devNum }
    
    resDetected <- resAllEvents <- NULL
    cntPlotsCNVstrong <- cntPlotsCNVweak <- cntPlotsUnclass <- cntPlotsSingleLoc <- iCnv <- 0  ;  iCan <- 111
    prevPlot <- FALSE
    
    #------------------- Main loop on candidates ------------------------------------
    for (iCan in 1:length(canMrkNames)) {

        canName <- as.vector(canMrkNames[iCan])
		if ( ! is.null(customCandidateList) && is.character(customCandidateList) &  ! canName %in% customCandidateList) next

        allCandMrksExceptIcan <- canMrkNames[canMrkNames != canName]
        segNotCan <- seg[ ! rownames(seg) %in% allCandMrksExceptIcan, ]   # keeps only Mendelian mrk (without any candidates) PLUS the marker iCan
        mapNotCan <- map[ ! map$snpName %in% allCandMrksExceptIcan, ]
        
        eD <- analyze_profiles(allChromLens=allChromLens, mapNotCan=mapNotCan, segNotCan=segNotCan, iCan=iCan, canName=canName, hypotheses=hypotheses
                          , minNumIndivHcand=limitminNumIndivForCurve, minNumIndivMcand=limitminNumIndivForCurve
                          , maxGenDistToFusePeaks=maxGenDistToFusePeaks
                          , minNbMrkPerTarget=minNbMrkPerTarget, toleranceFreq=toleranceFreq, minSlopePeak=minSlopePeak, minR2peak=minR2peak
                          , winSizeToFitPeakShape=winSizeToFitPeakShape, PLOT_REGRESSIONS=F)

        if ( ! is.null(eD$type$type)) {
            deviceToPlot <- c(eD$type$type %in% CNVstrong, eD$type$type %in% CNVweak, eD$type$type %in% "-", eD$type$type %in% singleLocusEvents)
            deviceToPlot <- devList[deviceToPlot]
        } else {
            deviceToPlot <- NULL
        }
        probRules <- list( pAh=0, pAm=0, pBh=0, pBm=0, pHa=0, pHb=0, pHm=0 )
        alleleFreq <- c(eD$type$trueLoci$afAAA[2], NA, NA, eD$type$trueLoci$afBBB[2], NA, NA)
        if ( sum(eD$type$scoreHyps)>0) indexHyp <- max(which(eD$type$scoreHyps)) else indexHyp <- NA
        if (indexHyp %in% 1:length(hypotheses)) {
            if (! is.null(err_funcs[[indexHyp]]$pAh)) probRules$pAh <- err_funcs[[indexHyp]]$pAh(eD$nbObs, af=alleleFreq)
            if (! is.null(err_funcs[[indexHyp]]$pAm)) probRules$pAm <- err_funcs[[indexHyp]]$pAm(eD$nbObs, af=alleleFreq)
            if (! is.null(err_funcs[[indexHyp]]$pBh)) probRules$pBh <- err_funcs[[indexHyp]]$pBh(eD$nbObs, af=alleleFreq)
            if (! is.null(err_funcs[[indexHyp]]$pBm)) probRules$pBm <- err_funcs[[indexHyp]]$pBm(eD$nbObs, af=alleleFreq)
            if (! is.null(err_funcs[[indexHyp]]$pHa)) probRules$pHa <- err_funcs[[indexHyp]]$pHa(eD$nbObs, af=alleleFreq)
            if (! is.null(err_funcs[[indexHyp]]$pHb)) probRules$pHb <- err_funcs[[indexHyp]]$pHb(eD$nbObs, af=alleleFreq)
            if (! is.null(err_funcs[[indexHyp]]$pHm)) probRules$pHm <- err_funcs[[indexHyp]]$pHm(eD$nbObs, af=alleleFreq)
        }
        
        #------------------ Plot CNVs or Unclassified events -----------------------------
        if (is.null(deviceToPlot)) { 
            symbol <- "." } else if (names(deviceToPlot) %in% c("CNVstrong","CNVweak")) {
                symbol <- "+" } else if (names(deviceToPlot)=="Unclassified") {
                    symbol <- "?" } else if (names(deviceToPlot)=="SingleLocus") symbol <- "^"
        if (length(eD$type$type)>0) {
            toPlot <-   ( (eD$type$type %in% CNVstrong) && plotCNVstrong ) ||
                        ( (eD$type$type %in% CNVweak) && plotCNVweak ) ||
                        ( (eD$type$type %in% singleLocusEvents) && plotSingleLocus ) ||
                    ( ( ! eD$type$type %in% c(singleLocusEvents,CNVstrong,CNVweak)) && plotUnclassified )
            toPlot <- toPlot && !is.null(eD$prof)
            if (eD$type$type %in% c(CNVstrong,CNVweak)) iCnv <- iCnv + 1
        } else toPlot <- FALSE
        if (toPlot) { 
            thisType <- eD$type$type
            if (names(deviceToPlot)=="CNVstrong")    { cntPlotsCNVstrong <- cntPlotsCNVstrong + 1 ; plot <- cntPlotsCNVstrong }
            if (names(deviceToPlot)=="CNVweak")      { cntPlotsCNVweak <- cntPlotsCNVweak + 1     ; plot <- cntPlotsCNVweak }
            if (names(deviceToPlot)=="Unclassified") { 
                cntPlotsUnclass <- cntPlotsUnclass + 1  ; plot <- cntPlotsUnclass 
                thisType <- "Unclassified"
            }
            if (names(deviceToPlot)=="SingleLocus")  { 
                cntPlotsSingleLoc <- cntPlotsSingleLoc + 1 ; plot <- cntPlotsSingleLoc
            }
            if ( ! prevPlot) main1 <- "\n" else main1 <- ""
            if (names(deviceToPlot)=="CNVstrong")    main1 <- paste0(main1,"CNV-Strong   Plot:")
            if (names(deviceToPlot)=="CNVweak")      main1 <- paste0(main1,"CNV-Weak     Plot:")
            if (names(deviceToPlot)=="Unclassified") main1 <- paste0(main1,"Unclassified Plot:")
            if (names(deviceToPlot)=="SingleLocus")  main1 <- paste0(main1,"Single-Locus Plot:")
            addSpaces <- 13 - length(unlist(strsplit(eD$type$type,"")))
            mulSp <- ""  ;  for (ii in 1:addSpaces) mulSp <- paste0(mulSp, " ")
            thisType <- paste0(eD$type$type,mulSp)
            addSpaces <- 21 - length(unlist(strsplit(canName,"")))
            mulSp <- ""  ;  for (ii in 1:addSpaces) mulSp <- paste0(mulSp, " ")
            canNameAdj <- paste0(canName,mulSp)
            
            main2 <- paste0(main1, plot,"   CNVs:",iCnv,"   Chr:",map[map$snpName==canName,]$chr, "   Type:", thisType
                                     ,"  Cand:",iCan," ",canNameAdj," ")
            main <- paste(main2,"    " , paste(paste(names(eD$nbObs), eD$nbObs,sep = ":"),collapse="  "))
            mapNotCan <- map[ ! map$snpName %in% canMrkNames[canMrkNames != canName], ]
            
            dev.set(deviceToPlot)
            profile_plot(myMap=mapNotCan, prof=eD$prof, smProf=eD$smProf, main=main, nbObs=eD$nbObs, canName=canName, trueLoci=eD$type$trueLoci, SMOOTH=TRUE)
            cat(main2)
            
        } else plot=NA
        
        #------------------ Simulations -----------------------------
        if(toPlot && !is.null(eD$prof) && simulateProfiles && eD$type$type %in% c(CNVstrong,CNVweak)) {
            cat("  Simulation" )
            typeSim <- eD$type$type
            refChro <- map[map$snpName==canName,]$chr
            refLoc <- eD$type$trueLoci[1, ]
            otherLoci <- eD$type$trueLoci[-1,] # ;  otherLoci <- otherLoci[order(otherLoci$iStart),]
            if(dim(otherLoci)[1]>=2){
                restPos <- c()  ;  restChr <- c()
                for (c in 1: dim(otherLoci)[1]) {
                    restPos <- c(restPos, mean(otherLoci[c,]$genStart, otherLoci[c,]$genStop) )
                    restChr <- c(restChr, otherLoci[c,]$chr)
                }
            }else{
                restPos <- mean(c(otherLoci$genStart, otherLoci$genStop))
                restChr <- otherLoci$chr
            }
            cntSim <- 0
            simOk <- FALSE
            while ( !simOk && cntSim < 5) {
                simSeg <- simulate_segregation_data(nuGamma=nuGamma, myMap=map, mySeg=seg, Fn=nbSelfingGenerationsRils, mappingPanelType=mappingPanelType
                                                , nbGenIntermat=nbIntermatingGenerations, scenario=typeSim, refMrkName=canName
                                                , otherLociChrs=as.vector(restChr), otherLociGenPos=as.vector(restPos)
                                                , fileName=NULL)
                                     #           , fileName=paste0(datePrefix,"_Simulation_Data_DEBUG"))  

                if ("|" %in% unlist(strsplit(eD$type$type, split=""))) {  ## case of systematic genotyping errors
                    simSeg[canName,] <- apply_systematic_errors(mrkData=simSeg[canName,], probRules=probRules)
                }

                simSegNotCan <- simSeg[ ! rownames(simSeg) %in% allCandMrksExceptIcan, ]
                classesToHide <- c()
                for (iClass in 1:(length(eD$nbObs)+1)) { 
                    if ( is.null(eD$prof$allFreqClass[[iClass]])) classesToHide <- c(classesToHide, iClass)
                }
                    
                sD <- analyze_profiles(allChromLens=allChromLens, mapNotCan=mapNotCan, segNotCan=simSegNotCan, iCan=iCan
                                  , canName=canName, hypotheses=hypotheses, minNumIndivHcand=10, minNumIndivMcand=10
                                  , maxGenDistToFusePeaks=maxGenDistToFusePeaks, minNbMrkPerTarget=minNbMrkPerTarget, toleranceFreq=toleranceFreq
                                  , minSlopePeak=minSlopePeak, minR2peak=minR2peak, winSizeToFitPeakShape=winSizeToFitPeakShape, PLOT_REGRESSIONS=FALSE)

                for (iClass in classesToHide) sD$prof$allFreqClass[[iClass]] <- NA
                simOk <- ( ! is.null(sD$prof))
                cntSim <- cntSim + 1
                if ( ! simOk) cat(paste0(" Skipped "))
            }
            if (length(sD$prof$allFreqClass)==7) {
				if (names(deviceToPlot)=="CNVstrong")    { cntPlotsCNVstrong <- cntPlotsCNVstrong + 1 ; simPlot <- cntPlotsCNVstrong }
				if (names(deviceToPlot)=="CNVweak")      { cntPlotsCNVweak <- cntPlotsCNVweak + 1     ; simPlot <- cntPlotsCNVweak }
				if (names(deviceToPlot)=="Unclassified") { cntPlotsUnclass <- cntPlotsUnclass + 1     ; simPlot <- cntPlotsUnclass }
				if (names(deviceToPlot)=="SingleLocus")  { cntPlotsSingleLoc <- cntPlotsSingleLoc + 1 ; simPlot <- cntPlotsSingleLoc }
				simMain <- paste0(main1, simPlot,"  SIMULATION    CNVs:",iCnv,"  Chr:",map[map$snpName==canName,]$chr, "  Type: "
								   , sD$type$type,"   Marker ",iCan,"  ",canName," ")
				simMain <- paste(simMain,"      " , paste(paste(names(sD$nbObs), sD$nbObs,sep = ":"),collapse="  "))
				
				dev.set(deviceToPlot)
                profile_plot(myMap=mapNotCan, prof=sD$prof, smProf=sD$smProf, main=simMain, nbObs=sD$nbObs
                         , canName=canName, trueLoci=sD$type$trueLoci, SMOOTH=TRUE)
            } 
        } 
        
        #------------------ Calculation of P-value -----------------------------
        if (toPlot && numberShufflesForPvalue > 0 && !is.null(eD$prof) && eD$type$type %in% c(CNVstrong,CNVweak)) {
            listTypesH0 <- c()
            if (plotShuffles) { 
				pdf(paste0(datePrefix,"_",canName,"_Plots_Shuffles_Pvalue.pdf"), 10, 1.5)
				par(cex=0.4)
			}
            cat("   P-val_H0=")
            nbMatchH0 <- 0

			if (nbCpus > 1) {
				allShuffles <- data.frame(matchH0=NULL, listTypesH0=NULL)
				registerDoParallel(nbCpus)
				allShuffles$listTypesH0 <- as.vector(allShuffles$listTypesH0)
				allShuffles <- foreach(iShufForPval = 1:numberShufflesForPvalue, .combine = rbind.data.frame) %dopar% {
					do_shuffle(seg, canName, eD, probRules, allCandMrksExceptIcan, allChromLens, mapNotCan, iCan, hypotheses
							, maxGenDistToFusePeaks, minNbMrkPerTarget, toleranceFreq, minSlopePeak, minR2peak, winSizeToFitPeakShape
							, singleLocusEvents, printDistribTypesShuffles, listTypesH0, plotShuffles, iCnv, iShufForPval, fileName=NULL )
		#					, fileName=paste0("Shuffle_Pvalue_Data_DEBUG",iShufForPval))  #  for debugging purposes
				}
				stopImplicitCluster()
			} else {
				allShuffles <- data.frame(matchH0=rep(NA,numberShufflesForPvalue), listTypesH0=rep(NA,numberShufflesForPvalue))
				for (iShufForPval in 1:numberShufflesForPvalue) {
					allShuffles[iShufForPval,] <- do_shuffle(seg, canName, eD, probRules, allCandMrksExceptIcan, allChromLens, mapNotCan, iCan, hypotheses
								, maxGenDistToFusePeaks, minNbMrkPerTarget, toleranceFreq, minSlopePeak, minR2peak, winSizeToFitPeakShape
								, singleLocusEvents, printDistribTypesShuffles, listTypesH0, plotShuffles, iCnv, iShufForPval, fileName=NULL )
		#				    	, fileName=paste0("Shuffle_Pvalue_Data_DEBUG",iShufForPval))  #  for debugging purposes
				}
			}
            pValueH0 <- sum( ! allShuffles$matchH0) / numberShufflesForPvalue
            pValueH0 <- round(pValueH0, ceiling(log(numberShufflesForPvalue)/log(10)))
            cat(pValueH0)
			if (printDistribTypesShuffles) { cat("\n") ; print(table(allShuffles$listTypesH0)) }
            if (plotShuffles) dev.off()
        } else pValueH0 <- NA
        
        #-------------------List all candidates with their identification status -----------
        if (is.null(eD$prof)) status <- "No_profile" else status <- eD$type$type
        if (status=="-") status <- "Unclassified"
        df <- data.frame(candidate=iCan, mrk=canName, status=status, plot=plot, Pval_H0=pValueH0
                         , pAh=round(probRules$pAh,2), pAm=round(probRules$pAm,2), pBh=round(probRules$pBh,2)
                         , pBm=round(probRules$pBm,2), pHa=round(probRules$pHa,2), pHb=round(probRules$pHb,2), pHm=round(probRules$pHm,2)
                         , AAA=eD$nbObs[1], AHA=eD$nbObs[2], AmA=eD$nbObs[3], BBB=eD$nbObs[4], BHB=eD$nbObs[5], BmB=eD$nbObs[6])
        resAllEvents <- rbind.data.frame(resAllEvents, df)
        write.table(resAllEvents, file=listFileName, quote=FALSE, row.names=FALSE, sep="\t")
        thisStats <- data.frame(Item=names(table(resAllEvents$status)), Value=as.numeric(table(resAllEvents$status)))
        listHyps <- as.vector(hypot$hypDisplay$type)
        listHyps <- c(listHyps, "No_profile","Unclassified")
        thisStatsAllHyps <- data.frame(Item=listHyps, Value=0)
        indices <- match(thisStats$Item, thisStatsAllHyps$Item)
        nas <- is.na(indices)
        thisStatsAllHyps[indices[!nas],]$Value <- as.numeric(table(resAllEvents$status)[!nas])
        statistics <- rbind.data.frame(firstStats,  thisStatsAllHyps)
        statistics$Value <- as.character(statistics$Value)
        write.table(statistics, file=statsFileName, quote=FALSE, row.names=FALSE, col.names=FALSE, sep="\t")
        
        #------------------ List identified CNVs with details of each locus -----------------------------
        if (!is.null(eD$prof) && eD$type$type %in% c(CNVstrong,CNVweak)) {  # if CNV identified then records event
            locus <- eD$type$trueLoci$locus
            chr <- eD$type$trueLoci$chr
            startPos <- eD$type$trueLoci$genPosInf
            stopPos <- eD$type$trueLoci$genPosSup
            mrkStart <- eD$type$trueLoci$mrkStart
            mrkStop <- eD$type$trueLoci$mrkStop
            nbLoc <- length(locus)
            df <- data.frame(signature=rep(names(deviceToPlot),nbLoc), plot=rep(plot,nbLoc), iCNV=rep(iCnv,nbLoc), iMrk=rep(iCan,nbLoc)
                             , type=rep(as.vector(eD$type$type),nbLoc)
                             , locus=locus, chr=chr, marker=rep(canName,nbLoc), mrkStart=mrkStart, mrkStop=mrkStop, startPos=startPos
                             , stopPos=stopPos, Pval_H0=pValueH0
                             , pAh=rep(round(probRules$pAh,2),nbLoc), pAm=rep(round(probRules$pAm,2),nbLoc), pBh=rep(round(probRules$pBh,2),nbLoc)
                             , pBm=rep(round(probRules$pBm,2),nbLoc), pHa=rep(round(probRules$pHa,2),nbLoc), pHb=rep(round(probRules$pHb,2),nbLoc)
                             , pHm=rep(round(probRules$pHm,2),nbLoc))
            resDetected <- rbind.data.frame(resDetected, df)
            write.table(resDetected, file=positionsFileName, quote=FALSE, row.names=FALSE, sep="\t")
        }
        prevPlot <- FALSE
        if (toPlot) { cat("\n") ; prevPlot <- TRUE } else cat(symbol)
            
    }  # End loop on iCan
    if (sum(!is.na(devList))>0) for (iDev in 1:sum(!is.na(devList))) dev.off()
    cat(paste0("\n",iCnv," CNVs detected\nAnalysis completed successfully :)\n"))
}








#' @title              Specifies CNV scenarii considered
#' 
#' @description Stores details of each hypothesis in a list. 
#'
#' @return list:\cr
#' $hyp list:\cr
#' \itemize{
#'    \item $hyp$type character string: describes the type of event: first locus 
#'     allele1/allele2, then "_", then idem for next locus, etc... (ex. "A/B_A/-")
#'    \item $hyp$sig character vector: describes the expected
#'     signature of the event in terms of peaks expected at each locus for the 
#'     different subsets of individuals in the population (ex. "A--BB-" "E--BA-"). 
#'     Each element of the vector corresponds to one locus. For each locus, this 
#'     pattern is encoded as a string of 6 characters, each of them corresponding to 
#'     one of the genotypic classes (1:AAA, 2:AHA, 3:A-A, 4:BBB, 5:BHB, 6:B-B). 
#'     For each of these 6 characters, "A" indicates a complete peak pointing upwards 
#'     to 100 percent frequency of allele "A", "B" indicates a complete peak pointing downwards 
#'     to 100 percent frequency of allele "B", "E" indicates a flat line around 50%, corresponding 
#'     to equal frequencies for "A" and "B" allele, "C" indicates a partial peak pointing 
#'     upwards to frequency of allele "A" between 50 percent and 100%, and "D" indicates a partial 
#'     peak pointing downwards to frequency of allele "B" between 50 percent and 100 percent
#' }\cr\cr
#' $hypDisplay data frame: same structure as $hyp, but $sig is a single string with 
#' each element of the vector separated by " | "\cr\cr
#' $estimate_errors list (over hypotheses 1 to 17) of lists of: $pAh, pAm, pBh, pBm, pHa
#' , pHb, pHm. Each such function has two parameters:\cr
#' - N vector [6] of numbers of individuals in the 6 3-locus classes
#' - af vector [6] numeric: allele frequencies for each class\cr
#' These functions return a list with each different probability pHm, pHa,etc...
#' 
#' @export
#' 
get_hypotheses =function() {
    
    #  Encoding used for signatures:
    #    CODE  PLANTS	      PEAKS          TREND    COMMENT
    #    -	   no	          NA	
    #    A	   yes	        A	
    #    B	   yes	        B	
    #    C	   yes	        no  	         A-like
    #    D	   yes	   	    no	           B-like
    #    E	   yes	        no	           flat
    #    a	   yes or no	  A or no (resp)	        peak dispensable (if no plants) but always complete
    #    b	   yes or no	  B or no (resp)          peak dispensable (if no plants) but always complete
    #    i	   yes or no	  yes or no	              ignore possible peaks
    
    # "type",	          "loc1",	   "loc2",    "loc3",	   "parentA",	"parentB",  "comment")
    hypo <- rbind(
        c("A/B",	      "A--B--",	 NA,	    NA,        "A",	     "B",        NA),  # 1
        c("A/B|Ah",	      "aA-B--",	 NA,	    NA,        "A",	     "B",        NA),  # 2
        c("A/B|Am",       "a-AB--",	 NA,	    NA,        "A",      "B",        NA),  # 3
        c("A/B|Bh",       "A--bB-",	 NA,	    NA,        "A",      "B",        NA),  # 4
        c("A/B|Bm",       "A--b-B",	 NA,	    NA,        "A",      "B",        NA),  # 5
        
        c("A/B_A/-",      "A-iBBi",  "E-iBAi",  NA,        "A",       "B",        NA),  # 6
        c("A/B_A/-|Hm",   "A-iB-B",  "E-iB-A",  NA,        "A",       "B",        NA),  # 7
        c("A/B_A/-|HmHb", "A-iBbB",	 "E-iDaA",  NA,        "A",       "B",        NA),  # 8
        
        c("A/B_-/A",      "A-iBBi",  "E-iABi",  NA,        "A",       "H",        NA),  # 9
        c("A/B_-/A|Hm",   "A-iB-B",  "E-iA-B",  NA,        "A",       "H",        NA),  # 10
        c("A/B_-/A|HmHb", "A-iBbB",  "E-iCbB",  NA,        "A",       "H",        NA),  # 11
        
        c("A/B_B/-",      "AAiB-i",  "BAiE-i",  NA,        "H",       "B",        NA),  # 12
        c("A/B_B/-|Hm",   "A-AB-i",  "B-AE-i",  NA,        "H",       "B",        NA),  # 13
        c("A/B_B/-|HmHa", "AaAB-i",  "DaAE-i",  NA,        "H",       "B",        NA),  # 14
        
        c("A/B_-/B",      "AAiB-i",  "ABiE-i",  NA,        "A",       "B",        NA),  # 15
        c("A/B_-/B|Hm",   "A-AB-i",  "A-BE-i",  NA,        "A",       "B",        NA),  # 16
        c("A/B_-/B|HmHa", "AaAB-i",  "CbBE-i",  NA,        "A",       "B",        NA),  # 17
        
 #       c("A/B_A/B",      "AAiBBi",  "ABiBAi",  NA,        "A",       "B",        NA),  # 18
 #       c("A/B_B/A",      "AAiBBi",  "BAiABi",  NA,        "H",       "H",        NA),  # 19
        
        c("A/B_-/A_-/A",    "A-iBBi",  "E-iADi",  "E-iADi",  "A",       "H",        NA),  # 18    # 20
        c("A/B_A/-_A/-",	"A-iBBi",  "E-iBCi",  "E-iBCi",  "A",       "B",        NA),  # 19    #  25
        c("A/B_-/B_-/B",	"AAiB-i",  "ADiE-i",  "ADiE-i",  "A",       "B",        NA),  # 20    #  30
        c("A/B_B/-_B/-",	"AAiB-i",  "BCiE-i",  "BCiE-i",  "H",       "B",        NA),  # 21    #  35
 
        c("A/B_-/A_A/-",    "A-iBBi",  "E-iADi",  "E-iBCi",  "A",       "H",        NA),  # 22    #  21
        c("A/B_A/-_-/A",	"A-iBBi",  "E-iBCi",  "E-iADi",  "A",       "H",        NA),  # 23    #  24
 
        c("A/B_-/A_-/B",    "AAiBBi",  "EEiABi",  "ABiEEi",  "A",       "H",        NA),  # 24    #  22
        c("A/B_-/B_-/A",	"AAiBBi",  "ABiEEi",  "EEiABi",  "A",       "H",        NA),  # 25    #  28

        c("A/B_-/A_B/-",	"AAiBBi",  "EEiABi",  "BAiEEi",  "H",       "H",        NA),  # 26    #  23
        c("A/B_B/-_-/A",	"AAiBBi",  "BAiEEi",  "EEiABi",  "H",       "H",        NA),  # 27    #  32
 
        c("A/B_A/-_-/B",	"AAiBBi",  "EEiBAi",  "ABiEEi",  "A",       "B",        NA),  # 28    #  26
        c("A/B_-/B_A/-",	"AAiBBi",  "ABiEEi",  "EEiBAi",  "A",       "B",        NA),  # 29    #  29
 
        c("A/B_A/-_B/-",	"AAiBBi",  "EEiBAi",  "BAiEEi",  "H",       "B",        NA),  # 30    #  27
        c("A/B_B/-_A/-",	"AAiBBi",  "BAiEEi",  "EEiBAi",  "H",       "B",        NA),  # 31    #  33
 
        c("A/B_-/B_B/-",	"AAiB-i",  "ADiE-i",  "BCiE-i",  "H",       "B",        NA),  # 32    #  31
        c("A/B_B/-_-/B",	"AAiB-i",  "BCiE-i",  "ADiE-i",  "H",       "B",        NA)   # 33    #  34
    )
    colnames(hypo) <- c("type","loc1","loc2","loc3","parentA","parentB","comment")
    hyp <- list()
    hypDf <- data.frame()
    nbLoc = nbPeaks = c()
    iHyp=1
    for (iHyp in 1:dim(hypo)[1]) {
        sigs = as.vector(hypo[iHyp,2:4])
        sigs = sigs[ !is.na(sigs) & sigs!="NA" & sigs!=""]
        nbLoc = c(nbLoc, length(sigs))
        nbPeaks = c(nbPeaks, sum(s2c(sigs) %in% c("A","B")))
        h = hypo[iHyp,]
        hyp[[iHyp]] = list(type=h["type"], parentA=h["parentA"], parentB=h["parentB"], signature=sigs, comment=h["comment"])
        df <- data.frame(type=h["type"], parentA=h["parentA"], parentB=h["parentB"], sig=paste(sigs, collapse=" | "))
        hypDf <- rbind.data.frame(hypDf, df)
    }
    
    estimate_errors <- list()
    estimate_errors[[1]]  <- list( pAh=NULL, pAm=NULL, pBh=NULL, pBm=NULL, pHa=NULL, pHb=NULL, pHm=NULL )
    estimate_errors[[2]]  <- list( pAh=function(N,af=NULL) N[2]/(N[1]+N[2]), pAm=NULL, pBh=NULL, pBm=NULL, pHa=NULL, pHb=NULL, pHm=NULL )
    estimate_errors[[3]]  <- list( pAm=function(N,af=NULL) N[3]/(N[1]+N[3]), pAh=NULL, pBh=NULL, pBm=NULL, pHa=NULL, pHb=NULL, pHm=NULL )
    estimate_errors[[4]]  <- list( pBh=function(N,af=NULL) N[5]/(N[5]+N[4]), pAh=NULL, pAm=NULL, pBm=NULL, pHa=NULL, pHb=NULL, pHm=NULL )
    estimate_errors[[5]]  <- list( pBm=function(N,af=NULL) N[6]/(N[6]+N[4]), pAh=NULL, pAm=NULL, pBh=NULL, pHa=NULL, pHb=NULL, pHm=NULL )
    
    estimate_errors[[6]]  <- list( pBm=NULL, pAh=NULL, pAm=NULL, pBh=NULL, pHa=NULL, pHb=NULL, pHm=NULL )
    
    estimate_errors[[7]]  <- list(    pHm=function(N,af=NULL)    N[6]/(N[5]+N[6])
                                    , pAh=NULL,pAm=NULL,pBh=NULL,pBm=NULL,pHa=NULL,pHb=NULL)

    estimate_errors[[8]]  <- list(    pHm=function(N,af=NULL) N[6]/(af[4]*N[4]+N[5]+N[6])
                                    , pHb=function(N,af=NULL) af[4]*N[4]/(af[4]*N[4]+N[5]+N[6])
                                    , pAh=NULL,pAm=NULL,pBh=NULL,pBm=NULL,pHa=NULL)
    
    estimate_errors[[9]]  <- list( pBm=NULL, pAh=NULL, pAm=NULL, pBh=NULL, pHa=NULL, pHb=NULL, pHm=NULL )
    
    estimate_errors[[10]]  <- list(  pHm=function(N,af=NULL)    N[6]/(N[5]+N[6])
                                   , pAh=NULL,pAm=NULL,pBh=NULL,pBm=NULL,pHa=NULL,pHb=NULL)
    
    estimate_errors[[11]] <- list(   pHm=function(N,af=NULL) N[6]/((1-af[4])*N[4]+N[5]+N[6])
                                   , pHb=function(N,af=NULL) (1-af[4])*N[4]/((1-af[4])*N[4]+N[5]+N[6])
                                   , pAh=NULL,pAm=NULL,pBh=NULL,pBm=NULL,pHa=NULL)
    
    estimate_errors[[12]]  <- list( pBm=NULL, pAh=NULL, pAm=NULL, pBh=NULL, pHa=NULL, pHb=NULL, pHm=NULL )
    
    estimate_errors[[13]] <- list(   pHm=function(N,af=NULL)    N[3]/(N[2]+N[3])
                                   , pAh=NULL,pAm=NULL,pBh=NULL,pBm=NULL,pHa=NULL,pHb=NULL)
    
    estimate_errors[[14]] <- list(   pHm=function(N,af=NULL) N[3]/(af[1]*N[1]+N[2]+N[3])
                                   , pHa=function(N,af=NULL) af[1]*N[1]/(af[1]*N[1]+N[2]+N[3])
                                   , pAh=NULL,pAm=NULL,pBh=NULL,pBm=NULL,pHb=NULL)
    
    estimate_errors[[15]]  <- list( pBm=NULL, pAh=NULL, pAm=NULL, pBh=NULL, pHa=NULL, pHb=NULL, pHm=NULL )
    
    estimate_errors[[16]] <- list(   pHm=function(N,af=NULL)    N[3]/(N[2]+N[3])
                                   , pAh=NULL,pAm=NULL,pBh=NULL,pBm=NULL,pHa=NULL,pHb=NULL)
    
    estimate_errors[[17]] <- list(   pHm=function(N,af=NULL) N[3]/((1-af[1])*N[1]+N[2]+N[3])
                                   , pHa=function(N,af=NULL) (1-af[1])*N[1]/((1-af[1])*N[1]+N[2]+N[3])
                                   , pAh=NULL,pAm=NULL,pBh=NULL,pBm=NULL,pHb=NULL)
    
    for (iH in 18:length(hyp)) estimate_errors[[iH]] <- list(pAh=NULL, pAm=NULL, pBh=NULL, pBm=NULL, pHa=NULL, pHb=NULL, pHm=NULL)
    
    return(list(hyp=hyp, hypDisplay=hypDf, estimate_errors=estimate_errors))
}






#' @title              Filter data
#' 
#' @description Filters out markers and plants based on multiple criteria (using thresholds):\cr
#' - Markers and plants with too many heterozygotes\cr
#' - Markers and plants with too many missing data\cr
#' - Markers and plants with min MAF too low
#'
#' @param map data frame: each row is a marker, columns are:
#' \itemize{
#'     \item $snpName: name of the marker
#'     \item $chr: chromosome (linkage group)
#'     \item $genPos: genetic position in cM of the marker
#' }
#' @param seg matrix: segregating genotypes coded as A, B, H, - Colnames: names of markers, 
#' rownames: names of individual in the population
#' @param maxPercentGenot Numeric: max percentage of one given genotype call
#' @param MAF numeric: Minimum Allele Frequency tolerated for considering
#' that segregation is Mendelian
#' @return list: \cr
#' $map (data frame). Each row is a marker, columns are:
#' \itemize{
#'     \item $snpName character string: name of the marker
#'     \item $chr character string or numeric: chromosome (linkage group)
#'     \item $genPos numeric: genetic position in cM of the marker
#' }
#' $seg matrix: segregating genotypes coded as A, B, H, - Colnames: names of markers, 
#' rownames: names of individual in the population
#' 
#' @examples
#' res <- filter_markers_plants(map=muMap, seg=mySeg, maxPercentGenot=90, minMAF=0.2)
#' seg = res$seg  ;  map = res$map
#' 
#' @export
#' 
filter_markers_plants = function(map, seg, maxPercentGenot , minMAF) {
    
    listMrk <- rownames(seg)
    # Restricts analysis to markers with positive genetic positions
    map <- map[map$genPos > 0,]
    listMrk <- listMrk[listMrk %in% map$snpName]
    map <- map[map$snpName %in% listMrk,]
    map <- map[order(map$genPos),]
    map <- map[order(map$chr),]
    seg <- seg[as.vector(map$snpName),]
    
    #  Filters out PLANTS with too many heterozygous markers
    apply(seg,2,function(x) sum(x == "H"))
    rem <- apply(seg,2,function(x) sum(x == "H")) > maxPercentGenot / 100 * dim(seg)[1]
    seg <- seg[, ! rem]
    cat(paste(sum(rem)," individual(s) removed because of having more than ",maxPercentGenot / 100 * dim(seg)[1], " H calls")) ; cat("\n")
    
    #  Filters out PLANTS with too many A  markers
    rem <- apply(seg,2,function(x) sum(x == "A")) > maxPercentGenot / 100 * dim(seg)[1]
    seg <- seg[, ! rem]
    cat(paste(sum(rem)," individual(s) removed because of having more than ",maxPercentGenot / 100 * dim(seg)[1], " A calls")) ; cat("\n")
    
    #  Filters out PLANTS with too many B markers
    rem <- apply(seg,2,function(x) sum(x == "B")) > maxPercentGenot / 100 * dim(seg)[1]
    seg <- seg[, ! rem]
    cat(paste(sum(rem)," individual(s) removed because of having more than ",maxPercentGenot / 100 * dim(seg)[1], " B calls")) ; cat("\n")
    
    #  Filters out MARKERS with too many missing data
    rem <- apply(seg,1,function(x) sum(x == "-")) > maxPercentGenot / 100 * dim(seg)[2]
    map <- map[ !rem,]
    seg <- seg[ ! rem,]
    cat(paste(sum(rem)," marker(s) removed because of having more than ",maxPercentGenot / 100 * dim(seg)[2]," missing data")) ; cat("\n")
    
    #  Filtering MARKERS with MAF < MIN_MAF
    nbA <- apply(seg, 1, function(x) sum(x == "A"))
    nbB <- apply(seg, 1, function(x) sum(x == "B"))
    nbNonMissing <- apply(seg, 1, function(x) sum(x != "-"))
    
    rem <- nbA < (minMAF * nbNonMissing) | nbB < (minMAF * nbNonMissing)
    map <- map[ ! rem,]
    seg <- seg[ ! rem,]
    cat(paste(sum(rem)," marker(s) removed because of having a Minimum AlleleFrequency lower than ",minMAF)) ; cat("\n") ; cat("\n")
    return(list(map=map, seg=seg))
}





#' @title              Select markers with H or Missing data
#' 
#' @description Returns candidate markers for being within CNV regions, defined 
#' as markers which have heterozygous and/or missing data for a minimum nmber of 
#' individuals in the population. \cr
#' Among non-candidate markers, filters out the non-candidate markers which show 
#' non-Mendelian segregation, based on the threshold value minMAF\cr
#' Reorders markers by chromosome and genetic position in seg and map tables
#' 
#' @param seg matrix: segregating genotypes coded as A, B, H, - Colnames: names of markers, 
#' rownames: names of individual in the population
#' @param map data frame: each row is a marker, columns are:
#' \itemize{
#'     \item $snpName character string: name of the marker
#'     \item $chr character string or numeric: chromosome (linkage group)
#'     \item $genPos numeric: genetic position in cM of the marker
#' }
#' @param minMAF numeric: minimum of Minimum Allele Frequency in the population 
#' for a marker to be considered in the analysis
#' @param minNumber numeric: minimum number of individuals Heterozygous (and/or Missing)
#' for the marker to be selected
#' @param selMrkWithHet logical: if TRUE, markers will be selected if they have at least 
#' minNumber individuals heterozygote
#' @param selMrkWithMissing logical: if TRUE, markers will be selected if they have at 
#' least minNumber individuals missing data
#' 
#' @return list:\cr
#' $seg matrix: same as param seg, but contains only (1) all candidate markers and 
#' (2) non-candidate Mendelian markers\cr
#' $map same as param map, but contains only (1) all candidate markers and 
#' (2) non-candidate Mendelian markers\cr
#' $canMrkNames vector of character strings: names of the candidate markers
#'
#' @export
#' 
mrk_with_het_or_missing <- function(seg, map, minMAF, maxNumIndivHmend, maxNumIndivMmend, minNumIndivHcand, minNumIndivMcand
                                    , selMrkWithHet=T, selMrkWithMissing=F){
    iCandMrk <- NULL
    if (minNumIndivHcand<1) minNumIndivHcand <- 1  ; if (minNumIndivMcand<1) minNumIndivMcand <- 1
    hasH <- apply(X=(seg == "H"), MARGIN=1, FUN=any)  	#  true if at least one plant is H for each marker
    hasM <- apply(X=(seg == "-"), MARGIN=1, FUN=any)  	#  true if at least one plant is MISSING for each marker
    if (selMrkWithHet) iCandMrk <- which(hasH)
    if (selMrkWithMissing) iCandMrk <- which(hasM)
    if (selMrkWithHet && selMrkWithMissing) iCandMrk <- which(hasM | hasH)
    nbHperMarker <- apply(seg,1,function(x) sum(x == "H"))
    nbMperMarker <- apply(seg,1,function(x) sum(x == "-"))
    isCandMrkForH <- as.vector(nbHperMarker >= minNumIndivHcand)
    isCandMrkForM <- as.vector(nbMperMarker >= minNumIndivMcand)
    
    if (selMrkWithHet) isCandMrk <- isCandMrkForH
    if (selMrkWithMissing) isCandMrk <- isCandMrkForM
    if (selMrkWithMissing && selMrkWithHet) isCandMrk <- isCandMrkForM | isCandMrkForH
    
    canMrkNames <- rownames(seg)[isCandMrk]
    # Among non-candidate markers, selects those with Mendelian segregation
    nonCandseg <- seg[ ! (isCandMrk),]
    fA <- as.vector(apply(nonCandseg,1,function(x) sum(x=="A")/sum(x %in% c("A","B"))))
    fB <- as.vector(apply(nonCandseg,1,function(x) sum(x=="B")/sum(x %in% c("A","B"))))
    isMendelianMaf <- fA >= minMAF & fA<=(1-minMAF) &  fB >= minMAF & fB<=(1-minMAF)
    isMendelianH <- as.vector(apply(nonCandseg,1,function(x) sum(x=="H")) <= maxNumIndivHmend) 
    isMendelianM <- TRUE  # as.vector(apply(nonCandseg,1,function(x) sum(x=="-")) <= maxNumIndivMmend) 
    isMendelian <- isMendelianMaf & isMendelianH & isMendelianM
    
    # Re-builds data sets with (1) all candidate markers and (2) non-candidate Mendelian markers
    newSeg = rbind(seg[isCandMrk,], nonCandseg[isMendelian,])
    newMap = map[map$snpName %in% rownames(newSeg),]
    newMap = newMap[order(newMap$genPos),]
    newMap = newMap[order(newMap$chr),]
    newSeg = newSeg[as.vector(newMap$snpName),]
    return(list(seg=newSeg, map=newMap, canMrkNames=canMrkNames))
}






#' @title              Assigns individuals to genotypic classes
#' 
#' @description Function for classifying each individual of the population 
#' into 16 genotypic classes depending on the values of the previous and next 
#' markers for each candidate marker.\cr
#' The 16 possible genotypic classes are: "AAA","AAB","ABA","ABB","AHA"
#' "AHB","A-A","A-B","BAA","BAB","BBA","BBB","BHA","BHB","B-A","B-B", with 
#' "XYZ" meaning X = marker before the marker under study, Y = marker under 
#' study, and Z = marker after the marker under study
#' 
#' @param snpName character string: name of the candidate marker under study
#' rownames: names of individual in the population
#' @param map data frame: each row is a marker, columns are:
#' \itemize{
#'     \item $snpName character string: name of the marker
#'     \item $chr character string or numeric: chromosome (linkage group)
#'     \item $genPos numeric: genetic position in cM of the marker
#' }
#' @param seg matrix: segregating genotypes coded as A, B, H, - Colnames: names of markers, 
#' rownames: names of individual in the population
#' 
#' @return list:\cr
#' $genotClass data frame where each row is one individual of the population (colanmes are 
#' individual names), with the following columns:
#' \itemize{
#'     \item $L: character: genotype at the left marker (before the candidate marker under study)
#'     \item $M: character: genotype at the candidate marker under study
#'     \item $R: character: genotype at the right marker (after the candidate marker under study)
#' }
#' $mrkLeft character string: name of the left marker
#' $mrkRight character string: name of the right marker
#' 
genotypic_classes <- function(snpName, map, seg) {
    
    ## return a data frame (plants classes index) containing the left and right genotype call for the candidate marker over all plants.
    chr <- map[which(map$snpName == snpName),]$chr
    mrkChr <- map[map$chr == chr,]
    segChr <- seg[as.character(mrkChr$snpName),]
    M <- which(mrkChr$snpName == snpName)
    
    mulNA <- rep(NA, dim(segChr)[2])
    genotClass <- data.frame(L=mulNA, M=mulNA, R=mulNA)
    genotClass$M <- as.character(segChr[M,])
    mrkLeft <- mrkRight <- mulNA
    iInd <- 1
    for (iInd in 1:dim(segChr)[2]) {
        if (M < dim(segChr)[1]) vec1 <- segChr[(M+1):dim(segChr)[1], iInd] else vec1 <- c()
        vec1 <- vec1 %in% c("A","B")
        if (sum(vec1)>0) {
            genotClass[iInd,]$R <- segChr[M + min(which(vec1)),iInd]
            mrkRight[iInd] <- rownames(segChr)[M + min(which(vec1))]
        } else  genotClass[iInd,]$R <- "-"
        if (M > 1) vec2 <- segChr[1:(M-1), iInd] else vec2 <- c()
        vec2 <- vec2 %in% c("A","B")
        if (sum(vec2)>0) {
            genotClass[iInd,]$L <- segChr[max(which(vec2)),iInd]
            mrkLeft[iInd] <- rownames(segChr)[max(which(vec2))]
        } else  genotClass[iInd,]$L <- "-"
    }
    return(list(genotClass=genotClass, mrkLeft=mrkLeft, mrkRight=mrkRight))
}





#' @title              Counts individuals in each genotypic class
#' 
#' @description Counting numbers of observations (individuals) in each genotypic class
#' 
#' @param $genotClass data frame where each row is one individual of the population (colanmes are 
#' individual names), with the following columns:
#' \itemize{
#'     \item $L: character: genotype at the left marker (before the candidate marker under study)
#'     \item $M: character: genotype at the candidate marker under study
#'     \item $R: character: genotype at the right marker (after the candidate marker under study)
#' }
#' @param S character: genotype code of the parent with the single copy (may be "A" or "B")
#' @param D character: genotype code of the parent with the single copy (may be "A" or "B")
#' @param I character: genotype code for missing data ("-")
#' @param H character: genotype code for (apparent) heterozygous genotype
#'    
#' @return vector: number of observations in each class. Names contain class names ("AAA",...)
#' 
count_obs_per_class <- function(genotClass, S, D, I, H) {
    
    nbObs <- c( sum(genotClass$L == S & genotClass$M == S & genotClass$R == S), sum(genotClass$L == S & genotClass$M == S & genotClass$R == D), sum(genotClass$L == S & genotClass$M == D & genotClass$R == S), sum(genotClass$L == S & genotClass$M == D & genotClass$R == D),
                sum(genotClass$L == S & genotClass$M == H & genotClass$R == S), sum(genotClass$L == S & genotClass$M == H & genotClass$R == D), sum(genotClass$L == S & genotClass$M == I & genotClass$R == S), sum(genotClass$L == S & genotClass$M == I & genotClass$R == D),
                sum(genotClass$L == D & genotClass$M == S & genotClass$R == S), sum(genotClass$L == D & genotClass$M == S & genotClass$R == D), sum(genotClass$L == D & genotClass$M == D & genotClass$R == S), sum(genotClass$L == D & genotClass$M == D & genotClass$R == D),
                sum(genotClass$L == D & genotClass$M == H & genotClass$R == S), sum(genotClass$L == D & genotClass$M == H & genotClass$R == D), sum(genotClass$L == D & genotClass$M == I & genotClass$R == S), sum(genotClass$L == D & genotClass$M == I & genotClass$R == D),
                sum(genotClass$L == H & genotClass$M == H & genotClass$R == H), sum(genotClass$L == H & genotClass$M == H & genotClass$R == S), sum(genotClass$L == H & genotClass$M == H & genotClass$R == D), sum(genotClass$L == S & genotClass$M == H & genotClass$R == H),
                sum(genotClass$L == D & genotClass$M == H & genotClass$R == H), sum(genotClass$L == H & genotClass$M == H & genotClass$R == I), sum(genotClass$L == I & genotClass$M == H & genotClass$R == H), sum(genotClass$L == I & genotClass$M == H & genotClass$R == I),
                sum(genotClass$L == S & genotClass$M == H & genotClass$R == I), sum(genotClass$L == D & genotClass$M == H & genotClass$R == I), sum(genotClass$L == I & genotClass$M == H & genotClass$R == S), sum(genotClass$L == I & genotClass$M == H & genotClass$R == D))
    classesNames <- c("AAA","AAB","ABA","ABB","AHA","AHB","A-A","A-B","BAA","BAB","BBA","BBB","BHA","BHB","B-A","B-B", "HHH","HHA", "HHB","AHH","BHH","HH-","-HH","-H-","AH-", "BH-","-HA","-HB")
    names(nbObs) <- classesNames
    return(nbObs)
}# End Function count_obs_per_class






#' @title              Compute genome-wide profiles
#' 
#' @description Calculates genome-wide allele frequency (1) among subsets of 
#' individuals corresponding to each genotypic class, and (2) among all 
#' individuals, as well as cumulated genetic positions genome-wide. \cr
#' Considers only the first 6 classes: AAA, AHA, A-A, BBB, BHB, B-B
#'  
#' @param myMap data frame: each row is a marker, columns are:
#' \itemize{
#'     \item $snpName character string: name of the marker
#'     \item $chr character string or numeric: chromosome (linkage group)
#'     \item $genPos numeric: genetic position in cM of the marker
#' }
#' @param mySeg matrix: segregating genotypes coded as A, B, H, - Colnames: names 
#' of markers, rownames: names of individual in the population
#' @param ixPerClass list: (one element per genotypic class) containing the 
#' indices of individuals belonging to the corresponding class
#' @param minNumIndiv numeric: minimum number of individuals per class for
#' which allele frequencies will be computed
#' 
#' @return list:\cr
#' $cumPos vector of numeric: cumulative gene positions over all chromosomes\cr
#' $allFreqClass list of vectors of numeric: each element of the list corresopnds 
#' to one of the 6 genotypic classes AAA, AHA, A-A, BBB, BHB, B-B, and contains 
#' a vector of frequencies of allele A among individuals belonging to the 
#' corresponding genotypic class. The values in the vector correspond to each 
#' marker along the genome. A last (7th) element corresponds to the vector of 
#' frequencies of allele A among all individuals, for each marker along the genome\cr
#' $allGenPos vector of numeric: genetic positions of markers genome-wide
#'  (chromosomes are one after each other but positions are not cumulated)\cr
#' $maxCumGenPos vector of numeric: genetic length of each chromosome 
#'  
genome_profiles <- function(allChromLens, myMap, mySeg, ixPerClass, minNumIndivHcand, minNumIndivMcand) {
    #'# --------- Private function ----------
    calc_freq <- function(genot) {
        freq <- as.vector(apply(genot,1,function(x) sum(x == "A")/(sum(x == "A")+sum(x == "B"))))
        freq[is.nan(freq)] <- 0.5
        return(freq)
    }
    #'# -------------------------------------
    allGenPos = cumPos = NULL
    allFreqClass <- vector("list",length(ixPerClass)+1)  ##  last index corresponds to ALL PLANTS
    chrs <- unique(myMap$chr)
    iChr <- 1
    for (iChr in 1:length(chrs)) {
        chr <- chrs[iChr]
        mrkChr <- myMap[myMap$chr == chr,]
        segChr <- mySeg[as.vector(mrkChr$snpName),]
        iClass <- 1
        for(iClass in 1:length(ixPerClass)){  ##	Classes: AAA, AHA, A-A, BBB, BHB, B-B
            ixPlant <- ixPerClass[[iClass]]
            if ( (iClass %in% c(1,4,2,5) & length(ixPlant) >= minNumIndivHcand) ||(iClass %in% c(3,6) & length(ixPlant) >= minNumIndivMcand))  {
     #           segChrThisClass <- data.frame(segChr[,ixPlant])
                segChrThisClass <- segChr[,ixPlant]
                allFreqClass[[iClass]] <- c(allFreqClass[[iClass]], calc_freq(segChrThisClass))
            }
        }
        freqTot <- calc_freq(segChr) # freq of A allele all classes together for this chrom
        allFreqClass[[length(ixPerClass)+1]] <- c(allFreqClass[[length(ixPerClass)+1]], freqTot)
        genPos <- mrkChr$genPos  #  genetic positions start already at zero, see load_data()
        allGenPos <- c(allGenPos, genPos)
        if (iChr > 1) cumPos <- c(cumPos, genPos + cumsum(allChromLens)[iChr-1] ) else cumPos <- genPos
    }
    return(list(cumPos=cumPos, allFreqClass=allFreqClass, allGenPos=allGenPos, maxCumGenPos=cumsum(allChromLens)))
}






#' @title              Fit peak flanks linearly
#' 
#' @description Fits points on each sides of a peak linearly over a window, and 
#' returns the R-squared and slope. Does this on both a window of size 
#' winSizeToFitPeakShape and a window of half that size, and keeps the highest 
#' R-squared
#'  
#' @param candPeak data frame: one row per candidate peak, as generated by 
#' the function find_peaks
#' @param myProf list: values of all allele frequencies, as generated by the 
#' function genome_profiles
#' @param iClass numeric: index of the genotypic class for which the peak has 
#' been detected (1:AAA. 2:AHA, 3:A-A, 4:BBB, 5:BHB, 6:B-B)
#' @param phase character: "A" for peak pointing to high frequency of allele 
#' "A" (upwards), or "B" for peak pointing to high frequency of allele "B"
#' (downwards)
#' @param myMap data frame: marker mapping information as generated by the
#' function load_map_file
#' @param winSizeToFitPeakShape numeric: length (in centiMorgans) on each 
#' side of the peak, where the shape will be fit to a line
#' @param PLOT logical: if TRUE, plots will be drawn for debugging purposes
#'     
#' @return list: list(worseR2=worseR2, bothSlopes=bothSlopes, bothPeakPos=bothPeakPos)
#' 
fit_peak_shape <- function(candPeak, myProf, iClass, phase, myMap, winSizeToFitPeakShape, PLOT_REGRESSIONS=FALSE) {
    
    plotLabel <- paste0("class ", candPeak$class,candPeak$phase," chr", candPeak$chr)
    mrkChr <- myMap[myMap$chr == candPeak$chr,]
    middlePeakPos <- (candPeak$genStart + candPeak$genStop)/2
    bothR2 <- bothSlopes <- bothPeakPos <- c()
    if (PLOT_REGRESSIONS) par(mfrow=c(2,2))
    side <- 1  ;  side <- -1
    for (side in c(-1,1)) {
        r2all <- r2small <- slopeAll <- slopeSmall <- peakPosAll <- peakPosSmall <- NA
        len <- winSizeToFitPeakShape
        #    if (side == -1) iBeg <- candPeak$iStart  else iBeg <- candPeak$iStop	
        if (side == -1) iBeg <- max(which(myMap$chr==candPeak$chr & myMap$genPos <= middlePeakPos - 0))
        if (side == 1)  iBeg <- max(which(myMap$chr==candPeak$chr & myMap$genPos <= middlePeakPos + 0))
        if (side == -1 && myMap[iBeg,]$genPos - min(mrkChr$genPos) < len)  len <- myMap[iBeg,]$genPos - min(mrkChr$genPos)
        if (side == 1  &&  max(mrkChr$genPos) - myMap[iBeg,]$genPos < len) len <- max(mrkChr$genPos) - myMap[iBeg,]$genPos
        if (side == -1) iEnd <- min(which(myMap$chr==candPeak$chr & myMap$genPos >= myMap[iBeg,]$genPos - len))
        if (side == 1)  iEnd <- max(which(myMap$chr==candPeak$chr & myMap$genPos <= myMap[iBeg,]$genPos + len))
        
        ratio <- 0.2
        iSmall <- max(which(myMap$chr==candPeak$chr & myMap$genPos <= (myMap[iBeg,]$genPos + (myMap[iEnd,]$genPos - myMap[iBeg,]$genPos) * ratio)))
        
        x <- side * (myProf$cumPos[iBeg:iEnd] - myProf$cumPos[iBeg])
        y <- myProf$allFreqClass[[iClass]][iBeg:iEnd]
        if (phase=="A") y = 1-y
        uni <- (diff(x)==0)
        x <- x[uni]  ;  y <- y[uni]
        
        if (length(unique(y)) >= 10 && max(x)>len/2) {
            regAll <- lm(y ~ 0 + x)
            slopeAll <- regAll$coefficients[1]
            r2all <- suppressWarnings(summary(regAll)$r.squared)
            InterceptRegAll <- lm(y ~ x)
            InterceptSlopeAll <- InterceptRegAll$coefficients[2]
            InterceptInterAll <- InterceptRegAll$coefficients[1]
            peakPosAll <- 0
            if (InterceptSlopeAll != 0) peakPosAll <- - InterceptInterAll / InterceptSlopeAll
            if (PLOT_REGRESSIONS) {  
                plot(x,y, cex=0.6, xlim=range(c(0,peakPosAll,x)), ylim=c(0,1), main=plotLabel, sub=paste0("R2 = ",r2all))
                abline(regAll, col=2); abline(InterceptRegAll, col=3)
                abline(v=peakPosAll, col=4, lwd=2); abline(h=0,lty=3); abline(v=0,lty=3)
            }
        } else if (PLOT_REGRESSIONS) plot(x,y)
        
        x2 <- y2 <- yr2 <- NULL  #   zoom on iSmall markers
        if (length(unique(y)) >= 10 && max(x)>len/2) {
            x2 <- side * (myProf$cumPos[iBeg:iSmall] - myProf$cumPos[iBeg])		## first half (in genetic length) only considered for reg2
            y2 <- myProf$allFreqClass[[iClass]][iBeg:iSmall]
        } else { y2 = 1 - y  ;  x2 = x  }
        if (phase=="A") y2 = 1-y2
        x2 <- unique(x2)  ;  y2 <- unique(y2)
        
        if (length(unique(y2)) >= 10  && max(x2)>len/2*ratio) {
            regSmall <- lm(y2 ~ 0 + x2)
            slopeSmall <- as.numeric(regSmall$coefficients[1])
            r2small <- suppressWarnings(summary(regSmall)$r.squared)
            InterceptRegSmall <- lm(y2 ~ x2)
            InterceptSlopeSmall <- as.numeric(InterceptRegSmall$coefficients[2])
            InterceptInterSmall <- as.numeric(InterceptRegSmall$coefficients[1])
            peakPosSmall <- 0
            if (InterceptSlopeSmall != 0) peakPosSmall <- - InterceptInterSmall / InterceptSlopeSmall
            if (PLOT_REGRESSIONS) {  
                plot(x2,y2, cex=0.6, xlim=range(c(0,peakPosSmall),x2), ylim=range(c(0,y2)), main=plotLabel, sub=paste0("R2 = ",r2small))
                abline(regSmall, col=2); abline(InterceptRegSmall, col=3)
                abline(v=peakPosSmall, col=4, lwd=2); abline(h=0,lty=3); abline(v=0,lty=3)
            }
        } else if (PLOT_REGRESSIONS) plot(x2,y2)
        best <- which.max(c(r2all, r2small))
        if (length(best)==0) best <- 1
        bothR2 <- c(bothR2, max(c(r2all, r2small)))
        bothSlopes <- c(bothSlopes, c(slopeAll, slopeSmall)[best])
        bothPeakPos <- c(bothPeakPos, c(peakPosAll, peakPosSmall)[best])
        bothPeakPos[abs(bothPeakPos)>5] <- NA
        
    }  # End for side
    
    names(bothR2) <- names(bothSlopes) <- c("Left", "Right")
    worseR2 <- suppressWarnings(min(c(bothR2), na.rm=T))
    bothPeakPos <- bothPeakPos[order(bothPeakPos, na.last=F)]
    bothPeakPos <- bothPeakPos + middlePeakPos
    
    return(list(worseR2=worseR2, bothSlopes=bothSlopes, bothPeakPos=bothPeakPos))
}







#' @title              Smooth allele frequency profiles
#' 
#' @description First computes a strong smoothing using splines to detect outliers
#' markers and remove them. Then uses the remaining markers for a much lighter
#' smoothing, which will be used for peak detection#' 
#' 
#' @param valuesX vector of numeric: marker cumulated positions
#' @param valuesY vector of numeric: cumulated allele frequencies to smooth
#' 
#' @return vector of numeric: smoothed allele frequencies for this chromosome
#' 
spline_smooth <- function(valuesX, valuesY) {
    PLOT <- FALSE
    if (PLOT) { plot(valuesX,valuesY) ;  abline(h=0)  ;  abline(h=1) }
    sm1 = smooth.spline(x=valuesX, y=valuesY, spar=0.7)#9)
    smy1 <- predict(sm1, valuesX)$y
    if (PLOT) points(valuesX,smy1, col=2, type ="l")
    res1=abs(smy1-valuesY)
    thres <- 0.2
    sel1 <- (res1 <= thres)
    if (PLOT) points(x=valuesX[sel1], y=valuesY[sel1], col=3)
    sm2 <- smooth.spline(x=valuesX[sel1], y=valuesY[sel1], spar=0.01)
    smy2 <- predict(sm2, valuesX)$y
    if (PLOT) points(x=valuesX, y=smy2, col=4 )
    return(smy2)
}






#' @title              Smooth genome-wide allele frequency profiles
#' 
#' @description Separates allele frequency profiles into single chromosomes and
#' calls the spline_smooth function to smooth them 
#' 
#' @param chromLengths vector of numeric: genetic lengths of each chromosome
#' @param valuesX vector of numeric: marker cumulated positions
#' @param valuesY vector of numeric: cumulated allele frequencies to smooth
#' 
#' @return vector of numeric: smoothed allele frequencies genome-wide
#' 
smooth_all_chroms <- function(chromLengths, valuesX, valuesY) {
    iChr <- 1
    smY <- c()
    for (iChr in 1:length(chromLengths)) {
        if (iChr > 1) chrBeg <- chromLengths[iChr-1] else chrBeg <- -1 
        chrEnd <- chromLengths[iChr]
        sel <- (valuesX > chrBeg & valuesX <= chrEnd)
        smY <- c(smY, spline_smooth(valuesX=valuesX[sel], valuesY=valuesY[sel]) )
    }
    return(smY)
}







#' @name find_peaks
#' @title              Detect and locate peaks of allele frequency 
#' 
#' @description Detects peaks having sufficient width and slopes on 
#' each side. Merges peaks too close to each other.\cr
#' Allele frequency profiles corresponding to each possible genotypic class 
#' will be analyzed one after each other for peaks pointing Up or Down. 
#'  
#' @param prof list: (as returned by function genome_profiles)\cr
#' @param myMap data frame: each row is a marker, columns are:
#' \itemize{
#'     \item $snpName character string: name of the marker
#'     \item $chr character string or numeric: chromosome (linkage group)
#'     \item $genPos numeric: genetic position in cM of the marker
#' }
#' @param minNbMrkPerTarget numeric: minimum number of markers within one 
#' target chromosome
#' @param maxGenDistToFusePeaks numeric: genetic distance in centiMorgans
#' below which two detected peaks will be considered as a single wide peak
#' @param freqThresh numeric: minimum allele frequency required to detect a peak
#' @param minSlopePeak numeric: minimum slope of the regression line on either 
#' side of the peak for the peak to be valid
#' @param winSizeToFitPeakShape numeric: minimum genetic distance in centiMorgans 
#' on each side of the peak, where allele frequencies will be fir by a regression 
#' line
#' 
#' @return list:\cr
#' $candPeaks: data frame (each row is a candidate peak), with columns:\cr
#' \itemize{
#'     \item $class numeric: index of the genotypic class (1:AAA. 2:AHA, 3:A-A, 
#'     4:BBB, 5:BHB, 6:B-B)
#'     \item $phase character: "A" for peak pointing to high frequency of allele 
#'     "A" (upwards), or "B" for peak pointing to high frequency of allele "B"
#'     (downwards)
#'     \item $chr character string or numeric: name of the chromosome where the 
#'     peak was detected
#'     \item $iStart numeric: index of the first marker of the peak
#'     \item $iStop numeric: index of the last marker of the peak
#'     \item $mrkStart character string: name of the first marker of the peak
#'     \item $mrkStop character string: name of the last marker of the peak
#'     \item $genStart numeric: genetic position in centiMorgan of the first 
#'     marker of the peak
#'     \item $genStop numeric: genetic position in centiMorgan of the last 
#'     marker of the peak
#'     \item $leftSlope  numeric: slope of the regression line on left side of the peak
#'     \item $rightSlope  numeric: slope of the regression line on right side of the peak
#'     \item $R2  numeric: R-squared of the linear regression on the peak side for 
#'     which the value is the smallest
#'     \item $leftPeakPos  numeric: smallest estimation of peak position
#'     \item $rightPeakPos  numeric: largest estimation of peak position
#' }\cr
#' $smProf list: allele frequenci profiles for each genotypic class as generated 
#' by the funciton genome_profile, but smoothed and filtered for outliers using splines. \cr
#'
find_peaks = function(prof, myMap, minNbMrkPerTarget, maxGenDistToFusePeaks,
                      freqThresh, minSlopePeak, minR2peak, winSizeToFitPeakShape, PLOT_REGRESSIONS=FALSE) {
    
    classesPhases = data.frame(class = c(1,1,2,2,3,3,4,4,5,5,6,6), phase = c("B","A","B","A","B","A","B","A","B","A","B","A"))
    candPeaks = data.frame()
    smProf <- prof
    for (iClass in 1:7) {   ##   Generates all smoothed profiles. "class" 7 corresponds to ALL PLANTS
        if (length(smProf$allFreqClass[[iClass]])>0) {
            chromLengths <- prof$maxCumGenPos
            valuesX <- prof$cumPos
            valuesY <- prof$allFreqClass[[iClass]]
            smProf$allFreqClass[[iClass]] <- freq <- smooth_all_chroms(chromLengths=chromLengths, valuesX=valuesX, valuesY=valuesY)
        }
    }
    k <- 9
    for (k in 1:dim(classesPhases)[1]) {
        iClass = classesPhases$class[k]
        phase = classesPhases$phase[k]
        detected = data.frame()
        freq <- smProf$allFreqClass[[iClass]]
        if (phase == "A") allPeaks = (freq >= 1 - freqThresh)
        if (phase == "B") allPeaks = (freq <= freqThresh)
        chrs = unique(myMap[allPeaks,]$chr)
        chrs = chrs[order(chrs)]
        if (length(chrs) > 0) {
            iChr <- 1
            for (iChr in 1:length(chrs)) {
                allChrs <- unique(myMap$chr)
                numChr <- which(allChrs %in% chrs[iChr])
                ix = which(myMap$chr == chrs[iChr])
                peak = allPeaks[ix]             # allPeaks[myMap$chr == chrs[iChr]]
                genePos = myMap[ix,]$genPos    # myMap[myMap$chr == chrs[iChr],]$genPos
                
                if (sum(peak) > 0) {
                    i = 1
                    while (i < sum(peak) && i < length(peak)-1) {
                        if (genePos[which(peak)[i+1]] - genePos[which(peak)[i]] < maxGenDistToFusePeaks) {
                            peak[which(peak)[i]:which(peak)[i+1]] = TRUE         #   for (j in which(peak)[i]:which(peak)[i+1]) peak[j] = TRUE
                        }
                        i = i + 1
                    }
                    ix = ix[-length(ix)]
                    peak[1] = FALSE
                    peak[length(peak)] = FALSE
                    peak1 = peak[-length(peak)]
                    peak2 = peak[-1]
                    peak1[is.na(peak1)] = FALSE
                    peak2[is.na(peak2)] = FALSE
                    left = ix[( ! peak1) & peak2] + 1
                    right = ix[peak1 & ( ! peak2)]
                    len1 = right - left + 1
                    ok = (len1 >= minNbMrkPerTarget)     ###  Filters out blocks with less than minNbMrkPerTarget markers
                    left = left[ok]
                    right = right[ok]
                    if (length(left) > 0) {
                        mrkLeft = as.character(myMap[left,]$snpName)  ;  mrkRight = as.character(myMap[right,]$snpName)
                        genStart = myMap[left,]$genPos     ;     genStop = myMap[right,]$genPos
                        mulNA = rep(NA, length(genStop))
                        cumPosBegin <- 0
                        if (numChr>1) cumPosBegin <- prof$maxCumGenPos[numChr-1]
                        df = data.frame(class=iClass, phase=phase, chr=chrs[iChr], iStart=left, iStop=right, mrkStart=mrkLeft
                                        , mrkStop=mrkRight, genStart=genStart, genStop=genStop
                                        , cumPosStart=genStart+cumPosBegin, cumPosStop=genStop+cumPosBegin
                                        #                           , leftSlope=mulNA, rightSlope=mulNA, R2=mulNA
                                        , leftPeakPos=mulNA, rightPeakPos=mulNA)
                        detected = rbind.data.frame(detected, df)
                    }
                }
            }
        }
        toRemove = NULL
        if (dim(detected)[1] != 0) {
            iPeak <- 1
            for ( iPeak in 1:dim(detected)[1]) {
                candPeak = detected[iPeak,]
                #        slope = fitPeakShape(candPeak=candPeak, prof=prof, iClass=iClass, phase=phase, myMap=myMap, winSizeToFitPeakShape=winSizeToFitPeakShape)
                myProf <- smProf
                shapeRet <- fit_peak_shape(candPeak=candPeak, myProf=myProf, iClass=iClass, phase=phase, myMap=myMap
                                           , winSizeToFitPeakShape=winSizeToFitPeakShape, PLOT_REGRESSIONS=PLOT_REGRESSIONS)
                #        detected[iPeak,]$leftSlope <- shapeRet$bothSlopes[1]
                #        detected[iPeak,]$rightSlope <- shapeRet$bothSlopes[2]
                detected[iPeak,]$leftPeakPos <- shapeRet$bothPeakPos[1]
                detected[iPeak,]$rightPeakPos <- shapeRet$bothPeakPos[2]
                #        detected[iPeak,]$R2 <- shapeRet$worseR2
                #        if (shapeRet$worseSlope <= minSlopePeak || shapeRet$worseR2 <= minR2peak) toRemove = c(toRemove, iPeak)
                if (shapeRet$worseR2 < minR2peak) toRemove = c(toRemove, iPeak)
            }
        }
        if (length(toRemove) > 0) detec = detected[ - toRemove,] else detec = detected
        if ( !is.null(detec)) candPeaks <- rbind(candPeaks, data.frame(detec))
    }
    candPeaks <- candPeaks[order(candPeaks$cumPosStart),]
    return(list(candPeaks=candPeaks, smProf=smProf))
}





#' @title              Checks presence of required classes and only them
#' 
#' @description Checks if the classes of individuals required by the
#' hypothesis are are present with numbers higher than thresholds
#'  
#' @param candPeaks data frame (each row is a candidate peak), with columns:
#' 
all_classes_present <- function(hyp, prof) {
    iLoc <- 1
    for (iLoc in 1:length(hyp)) {
        hypLoc <- unlist(strsplit(hyp[iLoc], ""))
        for (iClass in 1:length(hyp[[1]])) {
            if (hypLoc[iClass] == "-" &&  ! is.null(prof[[iClass]])) return(FALSE)
            if (hypLoc[iClass] %in% c("A","B","C","D","E") &&  is.null(prof[[iClass]])) return(FALSE)
        }
    }
    return(TRUE)
}





#' @title              Groups peaks that share a region
#' 
group_peaks_by_loci <- function(candPeaks) {
    iGroup <- 0
    mulNa <- rep(NA, dim(candPeaks)[1])
    groups <- cbind.data.frame(group=mulNa, candPeaks)
    iPeak1 <- 1
    for (iPeak1 in 1:(dim(candPeaks)[1])) {
        if (is.na(groups[iPeak1,]$group)) {
            iGroup <- iGroup + 1
            groups[iPeak1,]$group <- iGroup
            region <- c(groups[iPeak1,]$cumPosStart, groups[iPeak1,]$cumPosStop)
            for (iPeak2 in iPeak1:dim(groups)[1]) {
                if (is.na(groups[iPeak2,]$group)) {
                    if (region[1] <= groups[iPeak2,]$cumPosStop && region[2] >= groups[iPeak2,]$cumPosStart) {
                        region <- c(max(c(region[1],groups[iPeak2,]$cumPosStart)), min(c(region[2],groups[iPeak2,]$cumPosStop)))
                        groups[iPeak2,]$group <- iGroup
                    }
                }
            }
        }
    }
    return(groups)
}






#' @title              Checks compatibility of peak list at a locus
#' 
#' @description Determines if the observed list of peaks at a given locus
#' is compatible with the corresponding expected list of peaks
#' 
#' @param expLoc vector of 6 characters: each character corresponds to the 
#' expected peak for the corresponding class. Expected peaks are encoded as
#' A, B, C, D, E, -, a, b, i
#' @param obsLoc vector of 6 characters: each character corresponds to the 
#' observed peak for the corresponding class. Observed peaks are encoded using 
#' only A, B, or -
#' 
#' @return logical
#' 
compatible_peaks <- function(expLoc, obsLoc) {
    iClass <- 1
    for (iClass in 1:length(expLoc)) {
        if (expLoc[iClass]=="A" && obsLoc[iClass] != "A")                return(FALSE)
        if (expLoc[iClass]=="B" && obsLoc[iClass] != "B")                return(FALSE)
        if (expLoc[iClass]=="C" && obsLoc[iClass] != "-")                return(FALSE)
        if (expLoc[iClass]=="D" && obsLoc[iClass] != "-")                return(FALSE)
        if (expLoc[iClass]=="E" && obsLoc[iClass] != "-")                return(FALSE)
        if (expLoc[iClass]=="-" && obsLoc[iClass] != "-")                return(FALSE)
        if (expLoc[iClass]=="a" && ! obsLoc[iClass] %in% c("A","-"))     return(FALSE)
        if (expLoc[iClass]=="b" && ! obsLoc[iClass] %in% c("B","-"))     return(FALSE)
        if (expLoc[iClass]=="i" && ! obsLoc[iClass] %in% c("A","B","-")) return(FALSE)
    }
    return(TRUE)
}





#' @title              Get allele frequency
#' 
#' @description Returns allele frequency of the class at the marker position
#'  closest to pos
#' 
#' @param class numeric: index of the class (1 to 6)
#' @param pos numeric: genetic position in centiMorgan of the locu at which one
#' wants to measure allele frequency
#' @param prof profile as generated by function 'genome_profiles'
#' 
#' @return numeric
#' 
get_allele_freq <- function(class, cumPos, prof) {
    closestMrkIx <- which.min(abs(prof$cumPos - cumPos))
    freq <- prof$allFreqClass[[class]][closestMrkIx]
    freq <- max(freq, 0)
    freq <- min(freq,1)
    return(freq)
}






#' @title              Match hypothesis
#' 
#' @description Determines if a given type of structural variant is compatible with the 
#' list of peaks detected in the allele frequency profiles
#' 
#' @param candPeaks data frame (each row is a candidate peak), with columns:
#' \itemize{
#'     \item $class numeric: index of the genotypic class (1:AAA. 2:AHA, 3:A-A, 
#'     4:BBB, 5:BHB, 6:B-B)
#'     \item $phase character: "A" for peak pointing to high frequency of allele 
#'     "A" (upwards), or "B" for peak pointing to high frequency of allele "B"
#'     (downwards)
#'     \item $chr character string or numeric: name of the chromosome where the 
#'     peak was detected
#'     \item $iStart numeric: index of the first marker of the peak
#'     \item $iStop numeric: index of the last marker of the peak
#'     \item $mrkStart character string: name of the first marker of the peak
#'     \item $mrkStop character string: name of the last marker of the peak
#'     \item $genStart numeric: genetic position in centiMorgan of the first 
#'     marker of the peak
#'     \item $genStop numeric: genetic position in centiMorgan of the last 
#'     marker of the peak
#'     \item $slope  numeric: slope of the regression line on both sides of the peak
#' }
#' @param hyp vector of character strings: one element for each locus, contains 
#' the specification of the peak patterns expected for a given scenario of 
#' structural variant. For each locus, this pattern is encoded as a string 
#' of 6 characters, each of them corresponding to one of the genotypic classes (1:AAA,
#' 2:AHA, 3:A-A, 4:BBB, 5:BHB, 6:B-B). For each of these 6 characters, "A" indicates
#' a complete peak pointing upwards to 100 percent frequency of allele "A", "B" a complete 
#' peak pointing downwards to 100 percent of "B", "E" a flat line around 50 percent, (equal frequencies 
#' for "A" and "B" alleles, "C" a partial peak pointing upwards to 0.5 < f(A) < 1, and "D" 
#' a partial peak pointing downwards to 0.5 < f(B) < 1
#' @param myMap data frame: each row is a marker, columns are:
#' \itemize{
#'    \item $snpName: name of the marker
#'    \item $chr: chromosome (linkage group)
#'    \item $genPos: genetic position in cM of the marker
#' }
#' 
#' @return loci data frame: each row is a locus, columns are:\cr
#' \itemize{
#'    \item $locus numeric: index of the locus
#'    \item $chr numeric: chromosome (linkage group)
#'    \item $iStart numeric: index of the first marker within the region
#'    \item $iStop numeric: index of the last marker of the region
#'    \item $mrkStart character string: name of the first marker of the region
#'    \item $mrkStop character string: name of the last marker of the region
#'    \item $genPosInf numeric: genetic position (in centiMorgans) of the first 
#'    marker of the region
#'    \item $genPosMean numeric: middle genetic position (in centiMorgans) of the region
#'    \item $genPosSup numeric: genetic position (in centiMorgans) of the last 
#'    marker of the region
#' }
#' 
match_hypotheses <- function(candPeaks, hyp, myMap, nbObs, prof) {
    
    if ( ! all_classes_present(hyp, prof=prof)) return(data.frame())
    
    #  Form groups of peaks having some region in common
    groups <- group_peaks_by_loci(candPeaks=candPeaks)
    
    #  For each group, generates a signature corresponding to observed peaks
    observedSigs <- c()
    observedSigScalar <- rep("-", length(nbObs))
    iLoc <- 1
    for (iLoc in 1:max(groups$group)) {
        mySig <- observedSigScalar
        gr <- groups[groups$group==iLoc,]
        iClass <- 1
        for (iClass in 1:length(nbObs)) {
            grClass <- gr[gr$class==iClass,]
            if (dim(grClass)[1]==1) mySig[iClass] <- as.character(grClass$phase)
            if (dim(grClass)[1]>1) mySig[iClass] <- "-"  ##  forbidden situation (2 overlapping peaks of the same class in opposite directions)
        }
        observedSigs <- c(observedSigs, paste(mySig, collapse=""))
    }
    #  Checks compatibility between observedSigs and hyp
    if (length(hyp) != length(observedSigs)) return(data.frame())
    obsLocIndices <- rep(NA, length(hyp))
    iLoc <- 1
    for (iLoc in 1:length(hyp)) {
        expLoc <- unlist(strsplit(hyp[iLoc], ""))
        iObsLoc <- 1
        for (iObsLoc in rev(1:length(hyp))) {
            obsLoc <- unlist(strsplit(observedSigs[iObsLoc], ""))
            if (compatible_peaks(expLoc=expLoc, obsLoc=obsLoc) && ! iObsLoc %in% obsLocIndices) obsLocIndices[iLoc] <- iObsLoc
        }
    }
    if (sum(is.na(obsLocIndices)) > 0) return(data.frame())
    
    #  Summarizes information for each locus
    dfNa <- data.frame(locus=NA, chr=NA, iStart=NA, iStop=NA, mrkStart=NA, mrkStop=NA, genStart=NA, genStop=NA, genPosInf=NA, genPosMean=NA, genPosSup=NA)
    loci <- NULL
    iLoc <- 1
    for (iLoc in 1:length(hyp)) {
        df <- dfNa
        pks <- groups[groups$group==obsLocIndices[iLoc],]
        df$locus <- iLoc
        pks$chr <- as.vector(pks$chr)
        df$chr <- min(pks$chr)
        chrs <- unique(myMap$chr)
        iChr <- which(chrs %in% df$chr)
        #  if (min(pks$chr) != max(pks$chr)) stop("ERROR in match_hypotheses: different chromosomes found in the same 'group'locus'")
        df$iStart <- max(pks$iStart)
        df$iStop <- min(pks$iStop)
        df$mrkStart <- as.character(pks$mrkStart[which.max(pks$iStart)])
        df$mrkStop <- as.character(pks$mrkStop[which.min(pks$iStop)])
        df$genStart <- pks$genStart[which.max(pks$iStart)]
        df$genStop <- pks$genStop[which.min(pks$iStop)]
        df$genPosInf <- pks$genStart[which.max(pks$iStart)]
        df$genPosMean <- (pks$genStart[which.max(pks$iStart)] + pks$genStop[which.min(pks$iStop)])/2
        df$genPosSup <- pks$genStop[which.min(pks$iStop)]
        if (iChr==1) cumPos <- df$genPosMean else cumPos <- df$genPosMean + prof$maxCumGenPos[iChr-1]
        df$afAAA <- get_allele_freq(class=1, cumPos=cumPos, prof=prof)
        df$afBBB <- get_allele_freq(class=4, cumPos=cumPos, prof=prof)
        
        loci <- rbind.data.frame(loci, df)
    }
    if (sum(unlist(strsplit(hyp,""))=="|")>0) {  #  If systematic genotyping errors considered in the hypothesis
        
    }
    return(loci) 
}





#' @title              Identify CNV type based on profiles  
#' 
#' @description Compares the observed profile of allele frequencies with those 
#' expected under each of the hypotheses listed in the parameter "hypotheses"
#' 
#' @param candPeaks data frame: description of all peaks detected in the allele
#' frequency profiles, as returned by the function find_peaks
#' @param myMap data frame: genetic map as returned by the function load_data
#' @param nbObs vector of numeric: number of individuals assigned to each of the 
#' 6 genotypic classes given in names
#' @param prof list: allele frequencies for each marker and each genotypic class, as 
#' returned by the function genome_profiles
#' @param hypotheses list: description of all CNV hypotheses considered, as
#' returned by the function get_hypotheses
#' 
#' @return list:\cr
#' \itemize{
#'   \item $type character string: type of event (structural variant) identified: the different
#'   loci (duplicated regions) are separated by "_". Parental genotypes at each locus 
#'   are given as X/Y where X is the genotype of parent_1 and Y the genotype of parent_2
#'   \item $trueLoci data frame: each row is a locus (a copy of the CNV region).
#'   Columns are:\cr
#'   $hyp character string: type of event as defined above\cr
#'   $locus numeric: index of the locus (e.g. 1 or 2 for a 1:2 CNV)\cr
#'   $chr numeric: index of the chromosome (or linkage group)\cr
#'   $iStart numeric: index of the first marker of the region\cr
#'   $iStop numeric: index of the last marker of the region\cr
#'   $mrkStart character string: name of the first marker of the region\cr
#'   $mrkStop character string: name of the last marker of the region\cr
#'   $genStart numeric: genetic position of the first marker of the region\cr
#'   $genStop numeric: genetic position of the last marker of the region\cr
#'   \item $scoreHyps vector of logicals: each element is a type of event, 
#'   names are the event type as defined above, values are TRUE if the data 
#'   are compatible with the event considered, otherwise FALSE.
#' }
#'
CNV_type = function(candPeaks, myMap, nbObs, hypotheses, prof){ 
    
    scoreHyps <- rep(FALSE, length(hypotheses))
    names(scoreHyps) <- sapply(hypotheses, "[[", "type")
    allLoci <- NULL
    if (dim(candPeaks)[1] > 0) {
        iHyp <- 9
        for (iHyp in 1:length(hypotheses)) {
            hyp <- hypotheses[[iHyp]]$sig
            loci <- match_hypotheses(candPeaks=candPeaks,  hyp=hyp,  myMap=myMap,  nbObs=nbObs, prof=prof)
            
            if (dim(loci)[1] == 0) {
                candPeaksWoMd <- candPeaks[ ! candPeaks$class %in% c(3,6), ]  ##  Now tries again, but ignoring missing data
                if (dim(candPeaksWoMd)[1] > 0) {
                    loci <- match_hypotheses(candPeaks=candPeaksWoMd,  hyp=hyp,  myMap=myMap,  nbObs=nbObs, prof=prof)
                }
            }
            if (dim(loci)[1] > 0) {
                scoreHyps[iHyp] <- TRUE
                df1 <- cbind.data.frame(data.frame(hyp = rep(hypotheses[[iHyp]]$type, dim(loci)[1])), loci)
                allLoci <- rbind.data.frame(allLoci, df1)
            } else scoreHyps[iHyp] <- FALSE
        }
    }
    if (sum(scoreHyps) > 0) trueHyp = max(which(scoreHyps)) else trueHyp = 0   ## most complicated hypothesis
    if (trueHyp > 0) {
        type = hypotheses[[trueHyp]]$type
        sig = hypotheses[[trueHyp]]$sig
        trueLoci = allLoci[allLoci$hyp == hypotheses[[trueHyp]]$type,]
    } else {
        type = "-"
        sig = c()
        trueLoci = data.frame()
    }
    return(list(type=type, trueLoci=trueLoci, scoreHyps=scoreHyps))
}





#' @name limit_index_of_chroms
#' @title              Chromosome limits
#' 
#' @description returns genetic lengths of each chromosome
#' 
#' @param map data frame: as produced by load_data
#' 
#' @return vector of numeric: genetic lengths of each chromosome in centiMorgans
#' 
limit_index_of_chroms = function(map) {
    lims = c(1)
    for (chr in unique(map$chr)) {
        lims = c(lims, max(which(map$chr == chr)))
    }
    return(lims)
}





#' @title              Analyze allele frequency profiles
#'
#' @description Analyzes allele frequency profiles in sub-populations to
#' detect signatures of copy-number variation between the parents of the
#' mapping panel. Sub-populations are defined depending on the presence of 
#' heterozygous (and/or missing data) calls at the candidate marker locus.
#' Allele frequency profiles are generated, plotted, and signatures of CNVs 
#' are automatically detected and compared to a list of known signatures.
#'
#' @param map data frame: map information of all markers. Each row is a 
#' marker (markers are ordered by chromosome and genetic position), columns are:
#' \itemize{
#'     \item $map$snpName character string: name of the marker
#'     \item $map$chr character string or numeric: chromosome (linkage group)
#'     \item $map$genPos numeric: genetic position in cM of the marker
#' }
#' @param seg matrix: segregating genotypes coded as A, B, H, - Colnames: names 
#' of markers, rownames: names of individual in the population. Markers are 
#' ordered by chromosome and genetic position
#' @param iCan numeric: index of the candidate marker to be analyzed
#' @param canName character string: name of the candidate marker to be analyzed
#' @param hypotheses list: list of CNV scenarios considered, as generated by
#' the function get_hypotheses
#' @param minNumIndivHcand numeric: minimum number of Heterozygous plants with H call
#' @param minNumIndivMcand numeric: minimum number of Heterozygous plants with Missing call
#' 
#' @return list:
#' \itemize{
#'   \item $prof (as generated by function "genome_profiles")
#'   \item $allCandPeaks (as generated by function "find_peaks")
#'   \item $type (as generated by function "CNV_type")
#'   \item $nbObs (as generated by function "count_obs_per_class")
#' }
#' 
#' @export
#' 
analyze_profiles <- function(allChromLens, mapNotCan, segNotCan, iCan, canName, hypotheses, minNumIndivHcand
                             , minNumIndivMcand, maxGenDistToFusePeaks, minNbMrkPerTarget, toleranceFreq
                             , minSlopePeak, minR2peak, winSizeToFitPeakShape, PLOT_REGRESSIONS=FALSE) {
    
    gcRet <- genotypic_classes(snpName=canName, map=mapNotCan, seg=segNotCan)
    genoLMR <- gcRet$genotClass  ;  ML <- gcRet$mrkLeft  ;  MR <- gcRet$mrkRight
    
    nbObsPerClass <- count_obs_per_class(genotClass=genoLMR, S="A", D="B", I="-", H="H")
    interestingClasses <- c("AAA","AHA","A-A","BBB","BHB","B-B")
    interestingClassesIndex <- match(interestingClasses, names(nbObsPerClass))
    nbObs <- nbObsPerClass[interestingClasses]
    otherNbObs <- nbObsPerClass[-interestingClassesIndex]
    
    if (nbObs["AHA"] >= minNumIndivHcand || nbObs["BHB"] >= minNumIndivHcand 
        || nbObs["B-B"] >= minNumIndivMcand || nbObs["A-A"] >= minNumIndivMcand) {
        
        ixPerClass <- list()
        strLMR <- paste0(genoLMR$L, genoLMR$M, genoLMR$R)
        for (iClass in 1:length(nbObs)) ixPerClass <- c(ixPerClass, list(which(strLMR == names(nbObs[iClass]))))
        names(ixPerClass) <- names(nbObs)
        
        prof <- genome_profiles( allChromLens, myMap=mapNotCan, mySeg=segNotCan, ixPerClass=ixPerClass
                                , minNumIndivHcand=minNumIndivHcand, minNumIndivMcand=minNumIndivMcand )
        
        ret <- find_peaks( prof=prof, myMap=mapNotCan, minNbMrkPerTarget=minNbMrkPerTarget, maxGenDistToFusePeaks=maxGenDistToFusePeaks
                          , freqThresh=toleranceFreq, minSlopePeak=minSlopePeak, minR2peak=minR2peak, winSizeToFitPeakShape=winSizeToFitPeakShape
                          , PLOT_REGRESSIONS=PLOT_REGRESSIONS )
         
        candPeaks <- ret$candPeaks
        smProf <- ret$smProf
        
        type <- CNV_type( candPeaks=candPeaks, myMap=mapNotCan, nbObs=nbObs, hypotheses=hypotheses, prof=smProf )
        
    } else {
        prof <- candPeaks <- type <- smProf <- NULL
    }
    
    return(list(flankings=c(ML[1], MR[1]), prof=prof, candPeaks=candPeaks, type=type, nbObs=nbObs, smProf=smProf, otherNbObs=otherNbObs))  
}





#' @title              Plot profile 
#' 
#' @description Plots allele frequency profiles along the genome for different
#' subpopulations, together with the peaks detected
#' @param myMap data frame: marker mapping information as generated by function
#' load_data
#' @param prof list: values of allele frequencies for each marker and for each
#' subpopulation, as generated by function genome_profiles
#' @param main character string: title of the graph
#' @param nbObs named vector of numeric: numbers of individuals in each of the 
#' subpopulations, as generated by function count_obs_per_class. Names indicate
#' classes (e.g. AAA, AHA, etc...)
#' @param canName character string: name of the candidate marker for which 
#' heterozygous (and/or missing) data were observed
#' @param trueLoci data frame: information on the CNV event recognized by 
#' function CNV_type
#' @param supplLoci data frame: same as trueLoci, but for loci not recognized 
#' @param WIN_SIZE_PLOT numeric: size of the window (number of markers) for 
#' smoothing global allele frequencies (black line)
#' @param colVec vector of numeric: colors for the different classes of 
#' individuals
#' 
#' @export
#' 
profile_plot <- function(myMap, prof, smProf, main, nbObs, canName, SMOOTH=FALSE, trueLoci, supplLoci=NULL, WIN_SIZE_PLOT=10, colVec=c(2,6,8,4,5,1)) {
    
    nbChrs <- length(prof$maxCumGenPos)
    colVec=c(2,6,8,4,5,1)
    canChr <- myMap[myMap$snpName==canName,]$chr
    chrs <- unique(myMap$chr)
    iChr <- which(chrs==canChr)
    xlab = "Cumulated position"
    ylab = "Frequency of the P1 parent allele"
    plot(prof$cumPos, prof$allFreqClass[[7]], ylim=c(0,1), xlab=xlab, ylab=ylab, main=main, type="n", xaxt="n")
    xTicks <- seq(0, max(prof$cumPos), 20)
    axis(side=1, at=xTicks)
    iClass = 3
    for (iClass in c(3,6,2,5,1,4)) {
        if (length(prof$allFreqClass[[iClass]])==length(prof$cumPos)) {
            points(prof$cumPos, prof$allFreqClass[[iClass]], ylim=c(0,1), cex=0.15, xlab=xlab, ylab=ylab, col=colVec[iClass])
            if (SMOOTH) points(smProf$cumPos, smProf$allFreqClass[[iClass]], type="l", lwd=0.5, ylim=c(0,1), cex=0.4, xlab=xlab, ylab=ylab, col=colVec[iClass])
        }
    } 
    if (SMOOTH) points(smProf$cumPos, smProf$allFreqClass[[7]], type="l", lty=2, ylim=c(0,1), cex=0.3, xlab=xlab, ylab=ylab, col=1)
    if (is.null(supplLoci) || dim(supplLoci)[1]==0) {     #   adds unexpected detected peaks to graph
        legend(x="topleft", legend=names(nbObs), col=colVec, pch=1, cex=0.8, bty="n")
    } else {
        legend(x="topleft", legend=c(names(nbObs),"Suppl Peaks"), col=c(colVec, 7), pch=1, cex=0.8, bty="n")
    }
    abline(v=c(0,prof$maxCumGenPos), col=1, lwd=1.2)
    abline(h=c(0,1), lty=2)
    myMapBlock = myMap[myMap$chr==canChr,]
    blockCumPos = myMapBlock[myMapBlock$snpName==canName,]$genPos          #  - myMapBlock[1,]$genPos
    if (iChr > 1) blockCumPos = blockCumPos + prof$maxCumGenPos[iChr-1]
    mids = (prof$maxCumGenPos-c(0,prof$maxCumGenPos[-length(prof$maxCumGenPos)]))/2+c(0,prof$maxCumGenPos[-length(prof$maxCumGenPos)])
    text(mids, 0.1, sort(unique(myMap$chr)), col=1, cex=1.5)
    if ( ! is.null(trueLoci) && dim(trueLoci)[1]>0) {     #   adds identified peaks to graph
        iLoc = 1
        for (iLoc in 1:length(unique(trueLoci$locus))) {
            loci = trueLoci[trueLoci$locus==iLoc,]
            iSubLoc = 1
            for (iSubLoc in 1:dim(loci)[1]) {
                locus = loci[iSubLoc,]
                #      chrTarget = unique(myMap[locus$iStart,]$chr  )
                chrTarget <- unique(locus$chr)   #  should all be an the same chrom
                #      myMapTarget = myMap[myMap$chr %in% chrTarget,]
                #      targetCumPos = myMap[c(locus$iStart,locus$iStop),]$genPos - myMapTarget[1,]$genPos
                targetCumPos <- c(locus$genPosInf, locus$genPosSup)
                chrs <- unique(myMap$chr)   # is a scalar in fact
                iChr <- which(chrs %in% chrTarget)
                if (iChr > 1) targetCumPos = targetCumPos + prof$maxCumGenPos[iChr-1]
                color = iLoc
                rect(targetCumPos[1], 0.01, targetCumPos[2], 0.99, border=color, density=10, col=color, lwd=0.5)
            }
        }
    }
    #  if ( ! is.null(supplLoci) && dim(supplLoci)[1]>0) {     #   adds unexpected detected peaks to graph
    #    iLoc = 1
    #    for (iLoc in 1:dim(supplLoci)[1]) {
    #      locus = supplLoci[iLoc,]
    #      chrTarget = unique(myMap[locus$iStart,]$chr  )
    #      myMapTarget = myMap[myMap$chr %in% chrTarget,]
    #      targetCumPos = myMap[c(locus$iStart,locus$iStop),]$genPos   # - myMapTarget[1,]$genPos
    #      
    #      if (chrTarget > 1) targetCumPos = targetCumPos + prof$maxCumGenPos[chrTarget-1]
    #      color = 7   ###   Yellow
    #      rect(targetCumPos[1], 0.01, targetCumPos[2], 0.99, border=color, density=10, col=color, lwd=0.5)
    #    }
    #  }
}




#' @name gamma_meiosis
#' @title              Gamma_Meiosis
#' 
#' @description Simulates crossovers between parental haplotypes to generate
#' gamete haplotypes using the Gamma model for crossover interference
#' 
#' @param nuGamma = numeric: interference strength expressed as the nu parameter 
#' of the Gamma model. nu=1 for no interference. nu>1 for positive interference
#' @param chromosomeLength = numeric: genetic length (in centiMorgans) of the
#' chromosome considered
#' @param markerPosPerChromosome matrix: 1 x NbMarkers containing genetic positions 
#' of all markers 
#' @param chromosome matrix: 2 x NbMarkers with the diploid genotypes of parent
#' 
#' @return matrix: 1 x NbMarkers containing the gamete haplotype 
#'
#' @export
#' 
gamma_meiosis = function (nuGamma=1, chromosomeLength, markerPosPerChromosome, chromosome) {
    gamete = matrix (NA, ncol = length (markerPosPerChromosome [1,]), nrow = 1)
    positionCO = NULL
    posCO <- -10   #  Stationary Renewal Process starts at -10 Morgans to lose memory of the origin
    while (posCO < chromosomeLength/100) {  # chromLengths is in centiMorgans
        positionCO = c(positionCO,posCO)
        jumpLength = rgamma(1,shape=nuGamma,rate=2*nuGamma)  #  simulations are in Morgans
        posCO = posCO + jumpLength
    }
    positionCO <- positionCO[positionCO > 0]
    positionCO <- positionCO * 100  #  distances in centiMorgans
    vec01 <- NULL
    if (length(positionCO) != 0) { # thinning
        vec01 = sample(c(FALSE,TRUE), length(positionCO), replace=TRUE)
    }
    keepCO = c(0, positionCO[vec01], chromosomeLength)  #   forcing "a"false COs" at the extremities of the chrom
    homologue = sample(1:2,1) # parental chromosome selected at origin
    for (jLeft in 1:(length(keepCO)-1)) {
        Ix = which(keepCO[jLeft]<=markerPosPerChromosome[1,]  & markerPosPerChromosome[1,]<=keepCO[jLeft+1])
        if (length(Ix)>0) {
            gamete[1,Ix] = chromosome[homologue,Ix]
        }
        homologue = 3-homologue
    }
    return(gamete)
}




#' @name generate_gamete
#' @title              Simulate gamete
#' 
#' @description Simulate the genotype of a gamete, given that of the diploid parent
#' 
#' @param nbChroms numeric: number of chromsomes
#' @param chromLengths = vector of numeric: genetic length (in centiMoprgans) 
#' of each chromosome
#' @param markersPositions list: one element per chromosome. Each element is a 
#' matrix 1 x NbMarkers containing genetic positions of all markers
#' @param indiv list: (one element per chromosome) containing a matrix 
#' 2 x NbMarkers with the diploid genotypes of parent
#' 
#' @return list
#'       
generate_gamete = function(nuGamma, nbChroms, chromLengths, markersPositions, indiv) {
    gamete = vector("list", nbChroms)
    for (iChr in 1:nbChroms) {
        chromosome = indiv[[iChr]]
        markerPosPerChromosome = markersPositions[[iChr]]
        chromosomeLength = chromLengths[iChr]
        gamete[[iChr]] =  gamma_meiosis(nuGamma=nuGamma, chromosomeLength=chromosomeLength, markerPosPerChromosome=markerPosPerChromosome, 
                                        chromosome=chromosome)
    }
    return(gamete)
}





#' @name generate_offspring
#' @title              Simulate offspring diploid genotype
#' 
#' @description Creates genotypes of one offspring, given parents, for the 
#' different possible types of generation
#' 
#' @param nbChroms numeric: number of chromsomes
#' @param chromLengths = vector of numeric: genetic length (in centiMoprgans) 
#' of each chromosome
#' @param generationType character string: type of cross to be simulated,
#' must be one of "DH" (doubled haploids), "selfing", or "intercrossing"
#' @param markersPositions list: one element per chromosome. Each element is a 
#' matrix 1 x NbMarkers containing genetic positions of all markers
#' @param parentOne list: (one element per chromosome) containing a matrix 
#' 2 x NbMarkers with the diploid genotypes of parent_1
#' @param parentTwo list: (one element per chromosome) containing a matrix 
#' 2 x NbMarkers with the diploid genotypes of parent_1
#' 
#' @return list: 
#' 
generate_offspring <- function(nuGamma=1, nbChroms, chromLengths, generationType, markersPositions, parentOne, parentTwo) {
    genome <- vector("list", nbChroms)
    if(generationType == "DH") {
        indiv <- parentTwo
        gamete1 <- generate_gamete(nuGamma=nuGamma, nbChroms=nbChroms, chromLengths=chromLengths, markersPositions=markersPositions, indiv=indiv)
        gamete2 <- gamete1
    }
    if(generationType == "selfing") {
        indiv <- parentTwo
        gamete1 <- generate_gamete(nuGamma=nuGamma, nbChroms=nbChroms, chromLengths=chromLengths, markersPositions=markersPositions, indiv=indiv)
        gamete2 <- generate_gamete(nuGamma=nuGamma, nbChroms=nbChroms, chromLengths=chromLengths, markersPositions=markersPositions, indiv=indiv)
    }
    if(generationType == "intercrossing"){
        indiv1 <- parentOne
        indiv2 <- parentTwo
        gamete1 <- generate_gamete(nuGamma=nuGamma, nbChroms=nbChroms, chromLengths=chromLengths, markersPositions=markersPositions, indiv=indiv1)
        gamete2 <- generate_gamete(nuGamma=nuGamma, nbChroms=nbChroms, chromLengths=chromLengths, markersPositions=markersPositions, indiv=indiv2)
    }
    for (k in 1:nbChroms) {
        m = matrix(nrow=2, ncol=length(markersPositions[[k]][1,]))
        m[1,] <- gamete1[[k]][1,]
        m[2,] <- gamete2[[k]][1,]
        genome[[k]] <- m
    }
    return(genome)
}# End Function generate_offspring






#' @name generate_parents
#' @title              Simulate parental genotypes
#' 
#' @description Creates genotypes of parents, using alleles 0 and 1
#' 
#' @param nbIndiv numeric: number of parents to simulate (1 or 2)
#' @param nbChroms numeric: number of chromosomes
#' @param markersPositions list: one element per chromosome. Each element is a 
#' matrix 1 x NbMarkers containing genetic positions of all markers
#' 
#' @return list: one element per parent. Each element is a list (one element 
#' per chromosome) containing a matrix 2 x NbMarkers with the diploid genotypes 
#' set as 0, 1,... for each parent
#'
generate_parents = function(nbIndiv, nbChroms, markersPositions) {
    
    generation <- vector("list", nbIndiv)
    for (iInd in 1:nbIndiv) {
        genome <- vector("list", nbChroms)
        for (iChr in 1:nbChroms) {
            nbMrkrs <- length(markersPositions[[iChr]][1,])
            genome[[iChr]] <- matrix(NA, nrow=2, ncol=nbMrkrs)
            genome[[iChr]][1:2,] <- rep(iInd-1, 2*length(markersPositions[[iChr]][1,]))
        }
        generation[[iInd]] <- genome
    }
    return(generation)
}






#' @title              Simulate segregation data with CNVs
#' 
#' @description Simulates segregation data mimicking a data set, given a scenario
#' of copy-number variation, and generates corresponding .seg and .map files
#' 
#' @param myMap data frame: information about the marker data set to mimmic,
#' columns are:
#' \itemize{
#'     \item $snpName character string: name of the marker
#'     \item $chr character string or numeric: chromosome (linkage group)
#'     \item $genPos numeric: genetic position in cM of the marker
#' }
#' @param nbIndiv numeric: number of individuals in the population
#' @param Fn numeric: number of generations of selfing for Recombinant Inbred
#' Lines (RILs)
#' @param mappingPanelType character string: type of segregating population, must 
#' be one of: BC (backcross or doubled haploids population), RIL (recombinant inbred lines),
#' or IRIL (Intermated RILs), which means some generations of random intermating before the
#' generations of selfing
#' @param nbGenIntermate numeric: number of generations of random intermating in 
#' the case of IRILs
#' @param scenario character string: type of event (structural variant) to simulate:
#'  the different loci (duplicated regions) are separated by "_". Parental genotypes 
#'  at each locus are given as X/Y where X is the genotype of parent_1 and Y the 
#'  genotype of parent_2
#' @param refLocusMrkIndex numeric: index (as in the seg and map data files) of the marker 
#' considered as reference locus (first copy)
#' @param otherLociChrs vector of character strings: chromosomes (names) on which other
#' copies of the region will be simulated
#' @param otherLociGenPos vector of numeric: genetic position in centiMorgans of the 
#' other copies to be simulated
#' @param fileName character string: prefix of the names of the files that will be 
#' generated with the simulated segregation data
#' 
#' @return matrix of characters: each row is a marker and each column is an individual of the mapping 
#' population. Values are genotypes encoded with either "A", "B", or "H"
#' 
#' @export
#' 
simulate_segregation_data <- function(nuGamma, myMap, mySeg, Fn, mappingPanelType, nbGenIntermat, scenario, refMrkName, otherLociChrs, otherLociGenPos, fileName) {
    
    copies <- NULL
    nbIndiv <- dim(mySeg)[2]
    refMrkIndex <- which(myMap$snpName==refMrkName)
    myMap$snpName[refMrkIndex] <- paste("FirstCopy")
    iCo <- 1
    if (length(otherLociChrs)>0) for (iCo in 1:length(otherLociChrs)) {
        otherLoci <- myMap[refMrkIndex,] # only to take the structure of data
        otherLoci$snpName <- paste0("OtherCopy",iCo)
        otherLoci$chr <- otherLociChrs[iCo]
        otherLoci$genPos <- otherLociGenPos[iCo]
        copies <- rbind(copies,otherLoci )
    }
    myMap <- rbind(myMap, copies)
    myMap <- myMap[order(myMap$genPos),]
    myMap <- myMap[order(myMap$chr),]
    chrs <- unique(myMap$chr)
    nbChroms <- length(chrs)
    markersPositions <- vector("list", nbChroms)
    for (i in 1:nbChroms) markersPositions[[i]] <- t(as.matrix(myMap[myMap$chr == chrs[i],]$genPos))
    chromLengths <- as.vector(unlist(lapply(markersPositions, function(x) max(x))))
    mrkNames <- myMap$snpName
    
    parents <- generate_parents(nbIndiv=2, nbChroms=nbChroms, markersPositions=markersPositions) # Generate the parents
    
    nextGen <- vector("list", nbIndiv)
    # Generation F1
    for (iIndiv in 1:nbIndiv) nextGen[[iIndiv]] <- generate_offspring(nuGamma=nuGamma, nbChroms=nbChroms, chromLengths=chromLengths,   
                                                                      generationType="intercrossing", markersPositions=markersPositions, 
                                                                      parentOne=parents[[1]], parentTwo=parents[[2]])
    if (mappingPanelType == "DH") Fn <- 2
    #  Selfing generations F2 -> Fn
    iGenSelf <- 2
    for(iGenSelf in 2:Fn) { 
        generationFn = vector("list", nbIndiv)
        for (iIndiv in 1:nbIndiv) generationFn[[iIndiv]] <- generate_offspring(nuGamma=nuGamma, nbChroms=nbChroms, chromLengths=chromLengths,
                                                                               generationType="selfing", markersPositions=markersPositions,
                                                                               parentOne=NULL, parentTwo=nextGen[[iIndiv]])
        #  Random intermating generations for IRILs
        if(iGenSelf==2 && mappingPanelType=="IRIL"){
            for (iGenIntermating in 1:nbGenIntermat) { 
                generationFnIRIL= vector("list", nbIndiv)
                for (iIndiv in 1:nbIndiv) generationFnIRIL[[iIndiv]] <- generate_offspring(nuGamma=nuGamma, nbChroms = nbChroms, chromLengths = chromLengths,
                                                                                           generationType="intercrossing", markersPositions = markersPositions, 
                                                                                           parentOne = generationFn[[sample((1:nbIndiv)[-iIndiv],1)]], 
                                                                                           parentTwo = generationFn[[iIndiv]])
            }# End loop on iGenIntermating
            generationFn <- generationFnIRIL
        }# End if IRIL
        nextGen <- generationFn
    }# End loop on iGenSelf
    
    if(mappingPanelType != "DH"){
        FnCoding <- vector("list",nbIndiv)
        for (iIndiv in 1:nbIndiv) {
            for (iChr in 1:nbChroms) {
                indexA <- intersect(which(generationFn[[iIndiv]][[iChr]][1,]==0), which(generationFn[[iIndiv]][[iChr]][2,]==0 ))
                indexB <- intersect(which(generationFn[[iIndiv]][[iChr]][1,]==1), which(generationFn[[iIndiv]][[iChr]][2,]==1 ))
                indexH <- which(generationFn[[iIndiv]][[iChr]][1,]!=generationFn[[iIndiv]][[iChr]][2,])
                FnCoding[[iIndiv]][[iChr]] <- t(matrix(rep(NA,dim(generationFn[[iIndiv]][[iChr]])[2]), byrow = T))
                FnCoding[[iIndiv]][[iChr]][1,indexA] <- "A"
                FnCoding[[iIndiv]][[iChr]][1,indexB] <- "B"
                FnCoding[[iIndiv]][[iChr]][1,indexH] <- "H"
            }# End iChr
        }# End iIndiv
        dat <- NULL
        for(iChr2 in 1:nbChroms){
            dataPerChr <- NULL
            for(iInd2 in 1:nbIndiv){
                dataPerChr <- cbind(dataPerChr, FnCoding[[iInd2]][[iChr2]][1,])
            }
            dat <- rbind(dat, dataPerChr)
        }
    }# End if NOT DH
    
    if(mappingPanelType == "DH"){
        dat <- NULL
        for(iChr in 1:nbChroms){
            dataPerChr <- NULL
            for(iInd in 1:nbIndiv){
                dataPerChr <- cbind(dataPerChr, generationFn[[iInd]][[iChr]][1,])
            }
            dat <- rbind(dat,dataPerChr)
        }
        dat[dat==0] <- "A"
        dat[dat==1] <- "B"
    }# End if DH
    rownames(dat) <- mrkNames
    
    copy1Ix <- which(rownames(dat)=="FirstCopy")
    copy1 <- dat[copy1Ix,]
    if ("|" %in% unlist(strsplit(scenario, split=""))) {
        splitRes <- unlist(strsplit(scenario, split="|", fixed=TRUE))
        scenario <- paste(splitRes[1:(length(splitRes)-1)], collapse="")
    }
    iOtherLoc <- 1
    if (length(otherLociChrs)) for (iOtherLoc in 1:length(otherLociChrs)) {
        copy2andMoreIx <- which(rownames(dat)==paste0("OtherCopy",iOtherLoc))
        copy2andMore <- dat[copy2andMoreIx,]
        otherLocHyp <- unlist(strsplit(scenario, "_"))[iOtherLoc + 1]
        
        if(otherLocHyp=="A/-") dat[copy1Ix, copy1=="B" & copy2andMore=="A"] <- "H"
        if(otherLocHyp=="-/A") dat[copy1Ix, copy1=="B" & copy2andMore=="B"] <- "H"
        
        if(otherLocHyp=="B/-") dat[copy1Ix, copy1=="A" & copy2andMore=="A"] <- "H"
        if(otherLocHyp=="-/B") dat[copy1Ix, copy1=="A" & copy2andMore=="B"] <- "H"

        if(otherLocHyp=="A/B") dat[copy1Ix, (copy1=="A" & copy2andMore=="B") | (copy1=="B" & copy2andMore=="A")] <- "H"
        if(otherLocHyp=="B/A") dat[copy1Ix, (copy1=="A" & copy2andMore=="A") | (copy1=="B" & copy2andMore=="B")] <- "H"

        if(otherLocHyp=="A/B") dat[copy1Ix, (copy2andMore=="H")] <- "H"
        if(otherLocHyp=="B/A") dat[copy1Ix, (copy2andMore=="H")] <- "H"
    }
    
    rownames(dat)[which(rownames(dat)=="FirstCopy")] <- refMrkName
    myMap$snpName[which(myMap$snpName=="FirstCopy")] <- refMrkName
    colnames(dat) <- colnames(mySeg)
    if ( ! is.null(fileName)) {  # for debugging purposes only
        write.table(dat, paste(fileName,".seg",sep=""), quote=FALSE, row.names=TRUE, col.names=TRUE)
        write.table(myMap, paste(fileName,".map",sep=""), quote=FALSE, row.names=FALSE, col.names=FALSE)
    }
    
    if (length(otherLociChrs)>0) for (iOtherLoc in 1:length(otherLociChrs)) {  ## removes the fake additional markers from data
        dat <- dat[rownames(dat) != paste0("OtherCopy",iOtherLoc) , ]
    }
    dat <- dat[rownames(mySeg),]   ## orders dat according to initial mrk order in mySeg
    return(dat)
}




cat_load_bar <- function(min, max, value, char, width, label, newLineWhenMax=FALSE){
    min <- abs(as.integer(min)[1L])
    if( is.na(min) ){ stop("'min' should be an integer value.") }
    max <- abs(as.integer(max)[1L])
    if( is.na(max) ){ stop("'max' should be an integer value.") }
    if(min >= max){ stop("'min' should be strictly inferior to 'max.") }
    value <- abs(as.integer(value)[1L])
    if( is.na(value) ){ stop("'value' should be an integer value.") }
    if(value < min || value > max){ stop("'value' should be between 'min' and 'max' included.") }
    char <- as.character(char)[1L]
    if( !is1char(char) ){ stop("'char' should be a single character string.") }
    range <- max - min
    charFilled <- as.integer(value / range * width)
    charEmpty <- width - charFilled
    percent <- formatC( value / range * 100, format="f", digits=2 )
    cat(
        "|", rep(char, charFilled), rep(" ", charEmpty), "| "
        , percent, "% (", value, "/", range, " ", label, ")", 
        if( newLineWhenMax && value == max ){"\n"}else{"\r"}, 
        sep=""
    )
    invisible(NULL)
}




isHomozygous <- function(genotype){
    if( length(genotype) > 1L ){
        return( sapply(genotype, isHomozygous) )
    }
    if( !isTRUE(isGenotype(genotype)) ){
        stop("Invalid 'genotype' argument. Should be a 2 alleles genotype string.")
    }
    return( length(unique(s2c(genotype))) == 1L )
}




isHeterozygous <- function(genotype){
    return( !isHomozygous(genotype) )
}




is1char <- function(object){
    if( is.null(object) ){ return(FALSE) }
    if( length(object) != 1L ){ return(FALSE) }
    object <- as.character(object)
    if( is.na(object) ){ return(FALSE) }
    if( nchar(object) > 1L ){ return(FALSE) }
    return(TRUE)
}





isGenotype <- function(x){ 
    if( is.null(x) ){ return(FALSE) }
    if( is.list(x) ){
        return( rep(FALSE, length(x)) )
    }
    if( length(x) > 1L ){
        return( sapply(x, isGenotype) )
    }
    if( !length(x) ){ return(FALSE) }
    if( !is.character(x) ){ return(FALSE) }
    if( nchar(x) != 2L ){ return(FALSE) }
    return(TRUE)
}





isGenotypeM <- function(x, naChar="-"){ 
    if( !is1char(naChar) ){
        stop("Invalid parameter naChar. Should be a single character scalar.")
    }
    if( is.null(x) ){ return(FALSE) }
    if( is.list(x) ){
        return( rep(FALSE, length(x)) )
    }
    
    if( length(x) > 1L ){
        return( sapply(x, isGenotypeM) )
    }
    if( !length(x) ){ return(FALSE) }
    if( is.na(x) ){ return(FALSE) }
    if( !is.character(x) ){ return(FALSE) }
    if( nchar(x) %% 2 ){ return(FALSE) }
    x <- s2c(x)
    if( length(unique(x[x != naChar])) > 4L ){ return(FALSE) }
    return(TRUE)	
}





segregations <- function(p1, p2){
    if( !isTRUE(isGenotype(p1)) ){
        stop("Invalid 'p1' argument. Should be a 2 alleles genotype string.")
    }
    if( !isTRUE(isGenotype(p2)) ){
        stop("Invalid 'p2' argument. Should be a 2 alleles genotype string.")
    }
    # Vector of the parent's alleles:
    parentsAll <- c(s2c(p1), s2c(p2))
    return( sapply(
        list(
            c( parentsAll[1L], parentsAll[3L] ), # F0|M0
            c( parentsAll[1L], parentsAll[4L] ), # F0|M1
            c( parentsAll[2L], parentsAll[3L] ), # F1|M0
            c( parentsAll[2L], parentsAll[4L] )  # F1|M1
        ),
        c2s	) )
}




valid_descendants <- function( descendants, p1, p2, typePop, naChar="-" ){
    if( !is1char(naChar) ){
        stop("Invalid parameter naChar. Should be a single character scalar.")
    }
    if( !typePop %in% c("BC", "RIL", "IRIL") ){
        stop( "Invalid population type. Must be one of 'BC', 'RIL', 'IRIL'." )
    }
    if( !isTRUE(isGenotypeM(descendants)) ){
        if( !all(sapply(descendants, isGenotype)) ){
            stop( "Invalid argument 'descendants', should be a string with an ",
                  "even number of characters or a vector of 2 charcacter strings ", 
                  "with at most 4 unique characters overall." )
        }
    } else {
        descendants <- splitN(descendants, 2L)
    }
    # Relace 'missing data character' by '.' for use with grep (regex joker).
    descendants <- sapply( descendants, gsub, pattern=naChar, replacement="." )
    segs <- segregations( p1, p2 )
    segs <- segregations(segs[1L], segs[1L])
    if( typePop == "BC" ){
        desc_c <- table(descendants)
        segs <- c( 
            segs[ isHeterozygous(segs) ],
            names(sort(desc_c[ isHomozygous(names(desc_c)) ], decreasing=TRUE)[1L])
        )
    }
    segs  <- unique(unlist(c2s(c(s2c(segs), lapply( s2c(segs), "[", 2L:1L ) ))))
    return( structure( apply( 
        matrix( 
            unlist( lapply( 
                segs, 
                function(seg){ sapply(descendants, grepl, seg) } 
            ) ), 
            nrow=length(descendants) 
        ), 
        1L, any ), names=descendants ) )
}





s2c <- function(string) {
    # Transforms a strings into a vector.
    # Input: string [character scalar] string to split.
    # Output [character vector] => vector of characters composing the string.
    # WARNING: for string a length > 1 vector, all characters will be put in the same vector.
    string <- sapply( string, as.character )
    vect <- strsplit(string,"")
    if( length(vect) == 1){ vect <- unlist(vect) }
    return(vect)
}




splitN <- function(string, n){
    # Splits a string every n characters.
    # Input: string [character vector|scalar] => String(s) to split.
    # Input: n [integer scalar] => number of characters before each split.
    # Output: splitN [character vector|list[vector]] => result(s) of the string(s) splitting.
    n <- abs(as.integer(n)[1L])
    if( is.na(n) ){
        stop("Invalid 'n' parameter, should be a positive integer scalar.")
    }
    if( length(string) > 1L ){
        return( structure( lapply(string, splitN, n ), names=string ) )
    }
    string <- as.character(string)[1L]
    strsplit(string, paste0('(?<=', c2s(rep('.', n)), ')'), perl=TRUE)[[1L]]
}




c2s <- function(vect) {
    # Transforms a vector into a string.
    # Input: vect [vector] => vector to turn into string.
    # Output [character scalar] => resulting string.
    if( is.list(vect) ){
        string <- lapply(vect, paste, collapse="")
        if( length(vect) == 1L ){ string <- unlist(string) }
    } else {
        string <- paste(vect, collapse="")
    }
    return(string)
}





#' @title Determines if two genotypes are compatible
#' 
#' @name eq_genotypes
#' @description Takes into account possible missing data to determine whether 
#' or not two diploid genotypes are compatible
#' 
eq_genotypes <- function(g1, g2, naChar="-"){
    if( !is1char(naChar) ){
        stop("Invalid parameter naChar. Should be a single character scalar.")
    }
    if( !isTRUE(isGenotype(g1)) ){
        stop("Invalid 'g1' argument. Should be a 2 alleles genotype string.")
    }
    if( !isTRUE(isGenotype(g2)) ){
        stop("Invalid 'g2' argument. Should be a 2 alleles genotype string.")
    }
    g1 <- s2c(g1) ; g2 <- s2c(g2)
    g1 <- g1[g1 != naChar] ; g2 <- g2[g2 != naChar]
    # At least one of them has 2 missing alleles. They COULD be equivalent:
    if( any(c(!length(g1), !length(g2))) ){return(TRUE)}
    # They both have no, or a single missing allele, see if these are equivalent:
    if( length(g1) == length(g2) ){
        return( setequal(g1, g2) )
    } else { #One of them has a missing allele, they COULD be equivalent if the
        #	non missing allele is in the other genotype:
        return( if( length(g1) > length(g2) ){
            g2 %in% g1
        } else {
            g1 %in% g2
        } )
    }
}




#' @title Encode diploid genotypes in A B H - format
#' 
#' @name encode_genotype
#' @description Encode a genotype using either the letters encoding system or the hexadecimal encoding syetem.
#' @param p1 first parent's haplotype or vector of haplotypes. Length must be equal to p2's.
#' @param p2 second parent's haplotype or vector of haplotypes. Length must be equal to p1's.
#' @param descendants multi-genotype string or vector of multi-genotype strings.
#' @return encoded genotype(s) scalar/vector.
#' 
encode_genotype <- function(descendants, p1, p2, typePop, naChar="-", verbose=FALSE){
    if( !typePop %in% c("BC", "RIL", "IRIL") ){
        stop( "Invalid population type. Must be one of 'BC', 'RIL', 'IRIL'." )
    }
    if( !is1char(naChar) ){
        stop("Invalid parameter naChar. Should be a single character scalar.")
    }
    verbose <- isTRUE(as.logical(verbose)[1L])
    if( any( (lengths(list(descendants, p1, p2)) -> lens) > 1L ) ){
        if( length(unique(lens))-1L || any(lens < 1L) ){
            stop("Arguments 'descendants', 'p1' and 'p2' must have the same length.")
        }
        nCases <- length(descendants)
        if( verbose ){ 
            cat_load_bar( 0, nCases, 0, "\U25A0", 30, "encoded", TRUE)
        }
        encoding <- sapply(
            1L:nCases,
            function(iCase){
                enc <- encode_genotype(
                    descendants=descendants[iCase],
                    p1=p1[iCase],
                    p2=p2[iCase],
                    typePop=typePop,
                    naChar=naChar
                )
                if( verbose ){ 
                    cat_load_bar( 0, nCases, iCase, "\U25A0", 30, "encoded", TRUE)
                }
                return(enc) #Inner Return.
            } )
        return(encoding)
    }
    if( grepl(naChar, p1) | grepl(naChar, p2) ){
        # Incomplete parents, impossible to encode:
        return( c2s(rep("-", nchar(descendants)/2 )) )
    }
    descendants <- splitN(descendants, 2L)
    # Get the segregations of the two parents in the ?segregations order:
    segs <- segregations(p1, p2)
    segs <- segregations(segs[1L], segs[1L])
    if( (typePop == "BC") -> isBC ){
        desc_c <- table(descendants)
        segs <- c( 
            segs[ isHeterozygous(segs) ],
            names(sort(desc_c[ isHomozygous(names(desc_c)) ], decreasing=TRUE)[1L])
        )
    }
    names(segs) <- ifelse( isHomozygous(segs), 1L, 3L )
    if( !isBC ){
        names(segs)[ which(names(segs) == 1L)[2L] ] <- 2L
    }
    return( c2s( c("A", "B", "H", "-")[ as.integer( sapply(
        descendants,
        function(desc){
            let <- names(segs)[ 
                sapply(segs, eq_genotypes, desc) 
                ]
            return( if( setequal(c(1L, 3L), let) ){ # {'A', 'H'}
                4L
            } else if( setequal(c(2L, 3L), let) ){ # {'B', 'H'}
                4L
            } else if( setequal(c(1L:3L), let) ){ # {'A', 'B', 'H'}
                4L
            } else { unique(let) } )# {'A'} OR {'B'} OR {'H'} => itself.
        } ) ) ] ) )
}



#' @title Just a wrapper for the function encode_genotype
#' 
do_encoding <- function(segData, typePop, naChar="-", verbose=FALSE){
    segData$encoding <- encode_genotype(
        segData$genotype,
        segData$p1,
        segData$p2,
        typePop=typePop,
        naChar=naChar,
        verbose=verbose
    )
    return(segData)
}





#' @title              Load genetic map data (map)
#' 
#' @description Loads linkage map data. The working directory must contain 
#' a tab-delimited text file without headers with:\cr
#' - first column character string: marker names, \cr
#' - second column character string: chromosome, \cr
#' - third column numeric: genetic map position in centiMorgans.
#'
#' @param mapFileName character string: name of file containing genetic map 
#' positions
#' 
#' @return map data frame: each row is a marker, columns are:
#' \itemize{
#'     \item $snpName character string: name of the marker
#'     \item $chr character string or numeric: chromosome (linkage group)
#'     \item $genPos numeric: genetic position in cM of the marker
#' }
#' 
#' @examples
#' map <- load_map_data(mapFileName="myMapData.txt")
#' 
#' @export
#' 
load_map_file <- function(mapFileName) {
    
    map <- read.table(mapFileName, header=F, stringsAsFactors=FALSE)
    colnames(map)[1:3] <- c("snpName","chr","genPos")
    map <- map[order(map$genPos),]
    map <- map[order(map$chr),]
    map <- map[ ! is.na(map$chr),]
    map <- map[ ! is.na(map$genPos),]
    return(map)
}



#' @title             Load simple segregation matrix (seg)
#' 
#' @description Loads allele segregation data. The working directory must contain 
#' a tab-delimited text file named <segFileName> containing:\cr
#'  - A first line with all names of individuals in the population, then one 
#'    line per marker with\cr
#'  - Then one line per marker with marker name as first column and genotype 
#'    calls ("A", "B", "H", or "-") of each individual of the population in 
#'    subsequent columns, with the order of individuals corresponding to that
#'    of the first line
#'
#' @param segFileName characterstring: name of file containing the matrix of 
#' marker segregation data (encoded with A, B, H, -)
#' 
#' @return seg matrix: segregating genotypes coded as A, B, H, - Colnames: names 
#' of markers, rownames: names of individual in the population
#' 
#' @examples
#' seg <- load_seg_data(segFileName="mySegData.txt")
#' 
#' @export
#' 
load_seg_data <- function(segFileName) {
    seg <- read.table(segFileName, header=T, stringsAsFactors=FALSE, sep="")
    seg <- as.matrix(seg)
    return(seg=seg)
}



#' @title              Load Mapmaker-type segregation data (raw)
#' 
#' @description Loads allele segregation data from Mapmaker-type file (.raw)\cr
#' The working directory must contain a Mapmaker file named <segFileName> 
#' containing:\cr
#'  - a first line with 'data type <segType>', where <segType> can be one of 
#'  the four following: 'DH' (doubled haploids), 'RIL' (recombinant inbred 
#'  lines), or 'IRIL' (intermated recombinant inbred lines. In all cases, 
#'  allele segregation must be encoded with A, B, and H for residual/apparent 
#'  heterozygotes\cr
#'  - a second line with numbers of individuals, number of loci, and possible 
#'  aliases to be used for the segregation data as for the MapMaker software 
#'  (e.g. A=1 B=2 H=3 -=0)
#'  - a third line left blank
#'  - Then one line per marker with marker name as first column and genotype 
#'    calls ('A', 'B', 'H', or '-') of each individual of the population in 
#'    subsequent columns, with the order of individuals corresponding to that
#'    of the first line
#'
#' @param rawFileName characterstring: name of file containing segregation data
#' 
#' @return 
#' 
#' @export
#' 

load_raw_data <- function(rawFileName, sep="", indivNames=NULL){
    rawFileName <- as.character(rawFileName)[1L]
    if( !file.exists(rawFileName) ){ stop("File ", rawFileName, " not found.") }
    if( file_test( "-d", rawFileName) ){ stop(rawFileName, " is a directory.") }
    # Line by line reading for the header:
    encodingClasses <- c("A", "B", "H", "-")
    segDataConn <- file(rawFileName)
    open(segDataConn) # {
    on.exit(close(segDataConn)) # Auto-close file no matter function status.
    rgx_typePop <- " *data *type *"
    mrkName <- encoding <- aliases <- replacements <- character(0L) # Init.
    line <- readLines(segDataConn, n=1L)
    while( length(line) != 0L ){
        if( nchar(line) && !startsWith(line, '#') ){ # Not comment/empty line.
            if( startsWith(line, '*') ){ #Marker line:
                if( !nchar(sep) ){ sep <- "\\s+" }
                row <- strsplit( line, sep, perl=TRUE)[[1L]]
                mrkName <- c( mrkName, gsub("\\*", "", row[1L]) )
                encoding <- c(encoding, row[2L])
            } else { #Header line:
                if( grepl(rgx_typePop, line, ignore.case=TRUE) ){ # typePop header line:
                    typePop <- sub(rgx_typePop, "", line, ignore.case=TRUE)
                }
                if( grepl("symbols", line, ignore.case=TRUE) ){ #Aliases header line:
                    for( C in encodingClasses ){ #Iterate set of all posible codes.
                        if( any(unlist(regexec(paste0(C, "=(.)"), line) -> M ) > 0 ) ){ #If we find an alias for C.
                            aliases <- c(aliases, regmatches(line, M)[[1L]][2L])
                            replacements <- c(replacements, C)
                        }
                    }
                    sel <- aliases != replacements
                    aliases <- aliases[sel]
                    replacements <- replacements[sel]
                }
            }
        }
        line <- readLines(segDataConn, n=1L)
    }
    # }
    if( !typePop %in% c("DH", "RIL", "IRIL") ){
        stop( "Invalid population type. Must be one of 'DH', 'RIL', 'IRIL'." )
    }
    if( length(unique(nchar(encoding)))-1L ){
        stop("All markers must have the same number of individuals.")
    }
    nbIndiv <- nchar(encoding[1L])
    if( !is.null(indivNames) ){
        indivNames <- as.character(indivNames)[1L:nbIndiv]
        if( any(is.na(indivNames) | duplicated(indivNames) ) ){
            stop("Invalid 'indivNames' argument. Should be a vector of unique character strings.")
        }
    } else {
        indivNames <- paste0( "Ind", 1L:nbIndiv)
    }
    # Substitute aliases in encoding:
    encoding <- chartr( c2s(aliases), c2s(replacements), encoding )
    # Check for invalid encoding characters:
    codesInval <- ( unique(unlist(s2c(encoding))) -> codesUnq )[!codesUnq %in% encodingClasses]
    if( length(codesInval) != 0L ){ 
        stop("These invalid genotype codes: '", 
             paste( sort(unique(c( codesInval ))), collapse="', '"),
             "', may be present in the segregation data file (aliases and convertions applied).")
    }
    # Assemble data in dataframe:
    return(	structure( data.frame(
        mrkName=mrkName,
        p1=NA,
        p2=NA,
        genotype=NA,
        encoding=encoding,
        stringsAsFactors=FALSE
    ), typePop=typePop, indivNames=indivNames ) )
}



#' @title                Load SNP calling data (snp)
#' 
#' @name load_snp_data
#' @title Read .snp segregation file
#' @description Reads a .snp segregation file (Tabulated genotype file):
#'    \itemize{
#'        \item{ Column 1: }{ Marker names. Must not have duplicate names. }
#'        \item{ Columns 2 and 3: }{ Population parent's genotypes. Two characters per cell.
#'            Any characters may be used but must match with the characters used for the individuals. }
#'        \item{ Columns 4+: }{ Individuals' genotypes. Two characters per cell.
#'            Any characters may be used but must match with the characters used for the parents. }
#'    }
#' @param filename Name of the file to load.
#' @param typePop Population type code. Used to verify the validity of the segregations within the population.
#' @param popName Name for the population. Optional attribute that attaches to the returned dataframe.
#' @param indivNames Vector of individual names. Must be the same length as the number of individuals in the population.
#'    Must not have duplicate names. Optional attribute that attaches to the returned dataframe.
#' @param naChar Character used to represent missing data.
#' @param sep Character used to separate the table's cells.
#' @return Segregation Data dataframe of the population.
#' @export
load_snp_data <- function(snpFileName, typePop, indivNames=NULL, naChar="-", sep=""){
    filename <- as.character(snpFileName)[1L]
    naChar <- as.character(naChar)[1L]
    if( !is1char(naChar) ){
        stop("Invalid parameter naChar. Should be a single character scalar.")
    }
    if( !file.exists(snpFileName) ){ stop("File ", snpFileName, " not found.") }
    if( file_test( "-d", snpFileName) ){ stop(snpFileName, " is a directory.") }
    if( !typePop %in% c("DH", "RIL", "IRIL") ){
        stop( "Invalid population type. Must be one of 'DH', 'RIL', 'IRIL'." )
    }
    genTable <- read.table(
        snpFileName,
        sep=sep,
        numerals="no.loss",
        row.names=1L,
        stringsAsFactors=FALSE
    )
    if( !all(nchar(unlist(genTable)) == 2L)  ){
        stop("All genotypes should be two characters long.")
    }
    mrkName <- row.names(genTable)
    p1 <- genTable[,1L]
    p2 <- genTable[,2L]
    incParents <- grepl(naChar, p1) | grepl(naChar, p2)
    genotype <- sapply(
        1L:nrow(genTable),
        function(iMrk){
            c2s(as.character(genTable[iMrk,])[-c(1L:2L)])
        })
    # Check descendant validity:
    genotype_cut <- lapply(genotype[!incParents], splitN, 2L)
    # Detect genotypes which are invalid for descendants according to the population's parents:
    valid <- mapply( 
        valid_descendants, 
        genotype[!incParents], 
        p1[!incParents], 
        p2[!incParents], 
        MoreArgs=list(naChar=naChar, typePop=typePop), 
        SIMPLIFY=FALSE 
    )
    # Markers which have invalid genotypes. 
    whichHaveInvalid <- which( !sapply( valid, all) )
    # Replace the invalid genotypes with missing data characters:
    if( length(whichHaveInvalid) ){
        genotype[!incParents][whichHaveInvalid] <- sapply( mapply(
            function(ge, va){
                ge[!va] <- c2s(rep(naChar, 2L))
                return(ge)
            }, 
            genotype_cut[whichHaveInvalid], valid[whichHaveInvalid],
            SIMPLIFY=FALSE 
        ), c2s )
    }
    if( !is.null(indivNames) ){
        colnames(genTable)[-c(1L:2L)] <- indivNames
    }
    nbIndiv <- nchar(genotype[1L]) / 2
    if( colnames(genTable)[1L] == "V2" ){
        colnames(genTable)[-c(1L:2L)] <- paste0( "Ind", 1L:nbIndiv)
    }
    if( !is.null(indivNames) ){
        indivNames <- as.character(indivNames)[1L:nbIndiv]
        if( any(is.na(indivNames) | duplicated(indivNames) ) ){
            stop("Invalid 'indivNames' argument. Should be a vector of unique character strings.")
        }
    }
    return( structure(
        data.frame(
            mrkName=mrkName,
            p1=p1, p2=p2,
            genotype=genotype,
            encoding=NA,
            stringsAsFactors=FALSE ),
        typePop=typePop,
        indivNames=colnames(genTable)[-c(1L:2L)]
    ) ) 
}



#' @title    Load data for CNVmap
#' 
#' @description Master function to load data for CNVmap (depending on files 
#' present in the working directory). Loads allele segregation data under one of 
#' the possible file formats, and genetic map position data. The names of files 
#' to be read are:\cr
#' - mapFileName for mapping data, and\cr
#' - (rawFileName, snpFileName or segFileName, any one of these three files 
#' is sufficient) for segregation data.\cr
#' load_data keeps only markers present in both map and segregation files, and 
#' orders markers on chromosome and genetic position in both data sets.
#' 
#' @param mapFileName character string: file name for map data
#' @param segFileName character string: file name for encoded segregation data (matrix form)
#' @param rawFileName character string: file name for MapMaker-like raw data
#' @param snpFileName character string: file name for non-encoded segregation data (SNP alleles)
#' @param mappingPanelType character string: type of mapping population. May 
#' only be one of the following: "DH" (doubled haploid), "RIL" (Recombinant inbred
#' lines) or "IRIL" (intermated recombiant inbred lines)
#' 
#' @return list with two elements:\cr
#' $map data frame: each row is a marker (markers are ordered by chromosome 
#' and genetic position, columns are:
#' \itemize{
#'     \item $map$snpName character string: name of the marker
#'     \item $map$chr character string or numeric: chromosome (linkage group)
#'     \item $map$genPos numeric: genetic position in cM of the marker
#' }
#' $seg matrix: segregating genotypes coded as A, B, H, - Colnames: names 
#' of markers, rownames: names of individual in the population. Markers are 
#' ordered by chromosome and genetic position
#' 
#' @examples 
#' ldRet <- load_data(mapFileName="MyData.map", segFileName="MyData.seg", mappingPanelType="RIL")
#' seg <- ldRet$seg
#' map <- ldRet$map
#' 
#' @export
#' 
load_data <- function(mapFileName, segFileName, rawFileName=NULL, snpFileName=NULL, mappingPanelType) {
    
    if (file.exists(mapFileName) && ! file_test( "-d", mapFileName)) {
        map <- load_map_file(mapFileName)
    } else stop(paste0("couldn't find map file ", mapFileName))
    alleleDataLoaded <- "none"
    if (file.exists(segFileName) && ! file_test( "-d", segFileName)) {
        seg <- load_seg_data(segFileName=segFileName)
        alleleDataLoaded <- "seg"
    }
    if ( alleleDataLoaded == "none") {
        if (file.exists(rawFileName) && ! file_test( "-d", rawFileName)) {
            raw <- load_raw_data(rawFileName=rawFileName)
            seg <- do.call(rbind, s2c(raw$encoding) )
            colnames(seg) <- paste0("ind", 1:nchar(raw$encoding[1]))
            rownames(seg) <- raw$mrkName
            alleleDataLoaded <- "raw"
        }
    }
    if ( alleleDataLoaded == "none") {
        if (file.exists(snpFileName) && ! file_test( "-d", snpFileName)) {
            snp <- load_snp_data(snpFileName=snpFileName, typePop=mappingPanelType)
            snp <- do_encoding(snp, typePop=mappingPanelType)
            seg <- do.call(rbind, s2c(snp$encoding) )
            colnames(seg) <- paste0("ind", 1:nchar(snp$encoding[1]))
            rownames(seg) <- snp$mrkName
            
            alleleDataLoaded <- "snp"
        }
    }
    if ( alleleDataLoaded == "none") {
        stop(paste0("couldn't find any of the files ",
                    segFileName,", ",rawFileName,", or ",snpFileName))
    }
    mrkInBoth <- rownames(seg)[rownames(seg) %in% map$snpName]
    map <- map[map$snpName %in% mrkInBoth,]
    map <- map[order(map$genPos),]
    map <- map[order(map$chr),]
    for (chr in unique(map$chr)) {   #  Offsets the first genetic position to zero
        map[map$chr==chr,]$genPos <- map[map$chr==chr,]$genPos - min(map[map$chr==chr,]$genPos)
    }
    seg <- seg[as.vector(map$snpName),]
    return(list(seg=seg, map=map))
}




#' @title    Writes statistics on events found in a file
#' 
#' @description From the list of detected events as found in CNV_list, computes
#' summary statistics on the analysis
#' 
#' @param summary dataframe: with two columns:\cr
#' $item string: name of the parameter or category
#' $number numeric: value of the parameter, or number of events in the category
#' @param status vector of strings: contents the status of the event found for 
#' each candidate under the coded form of type e.g. A/B_A/-|Hm. 
#' @return dataframe: with two columns:\cr
#' $item string: name of the parameter or category
#' $number numeric: value of the parameter, or number of events in the category
#' 
write_stats_file <- function(statistics, status, file, maxNumIndivHmend, maxNumIndivMmend, maxNumIndivHcand, maxNumIndivMcand) {
    tab <- table(status)
    numTypes <- length(tab)
    return(rbind.data.frame(statistics, newDf))
}



#' @title    Applies systematic genotyping errors on data
#' 
#' @description From the list of probabilities of systematic errors, changes
#'  genotypes for one marker
#' 
#' @param mrkData vector of characters: contents the genotypes for a given marker
#' on all individuals
#' @param probeRules list of probabilities of systematic errors of the following
#' type:\cr
#' $pHm numeric: probability of turning "H" calls into missing data
#' $pHa numeric: probability of turning "H" calls into "A"
#' $pHb numeric: probability of turning "H" calls into "B"
#' @return dataframe: vector of characters: contents the genotypes for a given marker
#' on all individuals

apply_systematic_errors <- function(mrkData, probRules) {

    nbH <- sum(mrkData=="H")
    nbM <- round(probRules$pHm * nbH)
    nbA <- round(probRules$pHa * nbH)
    nbB <- round(probRules$pHb * nbH)
    while (nbA + nbB + nbM > nbH) nbM <- nbM - 1
    
    mrkData[mrkData=="H"][sample.int(n=nbH, size=nbM, replace=FALSE)] <- rep("-",nbM)
    nbH <- nbH - nbM
    mrkData[mrkData=="H"][sample.int(n=nbH, size=nbA, replace=FALSE)] <- rep("A",nbA)
    nbH <- nbH - nbA
    mrkData[mrkData=="H"][sample.int(n=nbH, size=nbB, replace=FALSE)] <- rep("B",nbB)
    return(mrkData)
}


#' @title    Shuffles data to remove CNV signature (H0 case)
#' 
#' @description Takes all individuals in which the ref mrk M* is in the A (resp B)
#' context and shuffles genotypes A, H, or - (resp B, H, or -) across those indiv
#' @param mySeg matrix: segregating genotypes coded as A, B, H, - Colnames: names 
#' of markers, rownames: names of individual in the population. Markers are 
#' ordered by chromosome and genetic position
#' @param refMrkName character string: name of the ML marker, which is Mendelian
#' and flanks M* on itl left to indicate the allelic context
#' @param leftMrkName character string: name of the MR marker, which is Mendelian
#' and flanks M* on itl right to indicate the allelic context
#' @param rightMrkName character string: name of the M* marker in which H or - calls
#' have been found in the data
#' @param fileName character string: name of the file in which the shuffled data
#' should be saved (optional, for debugging purposes)
#' @return matrix of characters: each row is a marker and each column is an individual of the mapping 
#' population. Values are genotypes encoded with either "A", "B", or "H"
#' 
#' @export
#' 
H0_shuffle_seq <- function(mySeg, refMrkName, leftMrkName, rightMrkName, fileName=NULL) {
    dat <- mySeg
    for (context in c("A","B")) {
        indices <- which(mySeg[leftMrkName,]==context & mySeg[rightMrkName,]==context)
        shufInd <- sample(x=indices, size=length(indices), replace=FALSE)
        dat[refMrkName , indices] <- mySeg[refMrkName , shufInd]
    }
    dat <- dat[rownames(mySeg),]   ## orders dat according to initial mrk order in mySeg
    if ( ! is.null(fileName)) {  # for debugging purposes only
        write.table(dat, paste0(fileName,".seg"), quote=FALSE, row.names=TRUE, col.names=TRUE)
    }
    return(dat)
}





#' @title    Does a shuffle for calculating the P-values
#' 
#' @description Shuffles the data, analyzes them, plots the profiles, and compares 
#' the predicted type with single-locus (H0)
#' @return data frame with columns: matchH0 (logical), listTypesH0 (string)
#' 
#' @export
#' 
do_shuffle <- function(seg, canName, eD, probRules, allCandMrksExceptIcan, allChromLens, mapNotCan, iCan, hypotheses
	, maxGenDistToFusePeaks, minNbMrkPerTarget, toleranceFreq, minSlopePeak, minR2peak, winSizeToFitPeakShape
	, singleLocusEvents, printDistribTypesShuffles, listTypesH0, plotShuffles, iCnv, iShufForPval, fileName=NULL) {
		
   shufSeg <- H0_shuffle_seq(mySeg=seg, refMrkName=canName, leftMrkName=eD$flankings[1]
				, rightMrkName=eD$flankings[2], fileName=fileName)
	           

	if ("|" %in% unlist(strsplit(eD$type$type, split=""))) {  ## case of systematic genotyping errors
		shufSeg[canName,] <- apply_systematic_errors(mrkData=shufSeg[canName,], probRules=probRules)
	}
	shufSegNotCan <- shufSeg[ ! rownames(shufSeg) %in% allCandMrksExceptIcan, ]
	classesToHide <- c()
	for (iClass in 1:(length(eD$nbObs)+1)) { 
		if ( is.null(eD$prof$allFreqClass[[iClass]])) classesToHide <- c(classesToHide, iClass)
	}
	pD <- analyze_profiles(allChromLens=allChromLens, mapNotCan=mapNotCan, segNotCan=shufSegNotCan, iCan=iCan
						   , canName=canName, hypotheses=hypotheses, minNumIndivHcand=10, minNumIndivMcand=10
						   , maxGenDistToFusePeaks=maxGenDistToFusePeaks, minNbMrkPerTarget=minNbMrkPerTarget
						   , toleranceFreq=toleranceFreq, minSlopePeak=minSlopePeak, minR2peak=minR2peak
						   , winSizeToFitPeakShape=winSizeToFitPeakShape, PLOT_REGRESSIONS=FALSE)
	matchH0   <- (length(pD$type$type)>0 && pD$type$type %in% singleLocusEvents)
	if ( ! is.null(pD$type$type) && printDistribTypesShuffles) {
		listTypesH0 <- c(listTypesH0, pD$type$type)
	} else listTypesH0 <- c(listTypesH0, "Unclassified")
	if (plotShuffles && length(pD$prof$allFreqClass)==7) {
		if (length(pD$prof$allFreqClass)==7) {
			pMain <- paste0("Plot: ",iShufForPval,"  Shuffles for Pvalue    CNVs:",iCnv,"  Chr:"
					,map[map$snpName==canName,]$chr, "  Type: ", pD$type$type,"   Marker ",iCan,"  ",canName," ")
			pMain <- paste(pMain,"      " , paste(paste(names(pD$nbObs), pD$nbObs,sep = ":"),collapse="  "))
			profile_plot(myMap=mapNotCan, prof=pD$prof, smProf=pD$smProf, main=pMain, nbObs=pD$nbObs
					 , canName=canName, trueLoci=pD$type$trueLoci, SMOOTH=TRUE)
		}
	}
	ret <- data.frame(matchH0=matchH0, listTypesH0=listTypesH0)
	ret$listTypesH0 <- as.vector(ret$listTypesH0)
	return(ret)
}
